from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows.jumplist import JumpListPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_os_windows_jumplist(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    custom_destination = absolute_path("_data/plugins/os/windows/jumplist/590aee7bdd69b59b.customDestinations-ms")
    automatic_destination = absolute_path("_data/plugins/os/windows/jumplist/5f7b5f1e01b83767.automaticDestinations-ms")

    user_details = target_win_users.user_details.find(username="John")

    fs_win.map_file(
        str(
            user_details.home_path.joinpath(
                "AppData/Roaming/Microsoft/Windows/Recent/CustomDestinations/590aee7bdd69b59b.customDestinations-ms"
            )
        )[3:],  # drop C:/
        custom_destination,
    )
    fs_win.map_file(
        str(
            user_details.home_path.joinpath(
                "AppData/Roaming/Microsoft/Windows/Recent/AutomaticDestinations/5f7b5f1e01b83767.automaticDestinations-ms"
            )
        )[3:],  # drop C:/
        automatic_destination,
    )

    target_win_users.add_plugin(JumpListPlugin)

    records = list(target_win_users.jumplist())
    assert len(records) == 3

    record = records[0]
    assert record.application_id == "590aee7bdd69b59b"
    assert record.application_name == "Powershell Windows 10"
    assert record.type == "customDestinations"



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from dissect.target.plugin import Plugin, arg, export
from dissect.target.plugins.filesystem.ntfs.mft import MftPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator


class MftTimelinePlugin(Plugin):
    """NTFS MFT timeline plugin."""

    def check_compatible(self) -> None:
        MftPlugin(self.target).check_compatible()

    @export(output="yield")
    @arg("--ignore-dos", action="store_true", help="ignore DOS file names")
    def mft_timeline(self, ignore_dos: bool = False) -> Iterator[str]:
        """Return the MFT records of all NTFS filesystems in a human readable format (unsorted) (deprecated, use mft.timeline).

        The Master File Table (MFT) contains metadata about every file and folder on a NFTS filesystem.

        If the filesystem is part of a virtual NTFS filesystem (a ``VirtualFilesystem`` with the MFT properties
        added to it through a "fake" ``NtfsFilesystem``), the paths returned in the MFT records are based on the
        mount point of the ``VirtualFilesystem``. This ensures that the proper original drive letter is used when
        available.
        When no drive letter can be determined, the path will show as e.g. ``\\$fs$\\fs0``.

        References:
            - https://docs.microsoft.com/en-us/windows/win32/fileio/master-file-table
        """  # noqa: E501
        warnings.warn(
            "The `mft_timeline` function is deprecated in favor of `mft.timeline` and will be removed in dissect.target 3.24",  # noqa: E501
            FutureWarning,
            stacklevel=2,
        )
        return MftPlugin(self.target).timeline(ignore_dos)



from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING

from defusedxml import ElementTree as ET

from dissect.target.filesystem import LayerFilesystem
from dissect.target.filesystems.zip import ZipFilesystem
from dissect.target.helpers import configutil
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path
    from uuid import UUID

    from dissect.target.target import Target

log = logging.getLogger(__name__)


@dataclass
class Extraction:
    type: str | None
    path: Path


@dataclass
class DeviceInfo:
    vendor: str
    model: str
    fguid: UUID | None = None
    guid: UUID | None = None
    os: str | None = None


@dataclass
class Ufdx:
    path: Path | None = None
    evidence: UUID | None = None
    device: DeviceInfo | None = None
    extractions: list[Extraction] | None = None


@dataclass
class Dump:
    type: str
    path: Path


@dataclass
class Keychain:
    type: str
    path: Path


@dataclass
class Ufd:
    path: Path
    device: DeviceInfo
    dumps: list[Dump] | None = None


class CellebriteLoader(Loader):
    """Load Cellebrite UFED exports (``.ufdx`` and ``.ufd``).

    References:
        - https://corp.digitalcorpora.org/corpora/mobile
    """

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)

        self.ufdx = None
        self.ufd = []
        self.ffs = None

        # Parse xml to find .ufd path
        if path.suffix == ".ufdx":
            try:
                tree = ET.fromstring(path.read_text())

                self.ufdx = Ufdx(
                    path=path,
                    evidence=tree.get("EvidenceID"),
                    device=DeviceInfo(**{k.lower(): v for k, v in tree.find("DeviceInfo").attrib.items()}),
                    extractions=[],
                )

                for extraction in tree.findall("Extractions/Extraction"):
                    self.ufdx.extractions.append(
                        Extraction(
                            type=extraction.get("TransferType"),
                            path=self.base_path.joinpath(extraction.get("Path").replace("\\", "/")),
                        )
                    )

            except ET.ParseError as e:
                raise ValueError(f"Invalid XML in {path}: {e}")

        elif path.suffix == ".ufd":
            self.ufdx = Ufdx(extractions=[Extraction(type=None, path=self.absolute_path)])

        else:
            raise ValueError(f"Unknown suffix {path.suffix} for {path}")

        # Not implemented: parse ufd ini to find ffs, could replace ``Extraction()`` with ``Ufd()``.

        for extraction in self.ufdx.extractions:
            if not extraction.path.is_file():
                log.warning("Extraction %s does not exist", extraction.path)
                continue

            config = configutil.parse(extraction.path, hint="ini")
            device = {k.lower(): v for k, v in config.get("DeviceInfo").items()}
            device["model"] += f" ({device['devicemodel']})"
            del device["devicemodel"]

            ufd = Ufd(
                path=extraction.path,
                device=DeviceInfo(**device),
                dumps=[],
            )

            for type, dump in config.get("Dumps").items():
                dump_path = ufd.path.resolve().parent.joinpath(dump)
                ufd.dumps.append(Dump(type=type, path=dump_path))

            self.ufd.append(ufd)

    @staticmethod
    def detect(path: Path) -> bool:
        return path.is_file() and path.suffix in [".ufdx", ".ufd"]

    def map(self, target: Target) -> None:
        for ufd in self.ufd:
            for dump in ufd.dumps:
                # Keychain dumps are inside the same FFS zip file, so we let the CellebriteFilesystem handle mounting
                # that so we prevent an extra file handle being opened.
                if dump.type != "FileDump":
                    log.warning("Ignoring Cellebrite dump %s of type %s", dump.path.name, dump.type)
                    continue

                if (size := dump.path.lstat().st_size) > 1_000_000_000:
                    log.warning(
                        "Cellebrite filesystem dump %s is %s GB, this might take a while..",
                        dump.path.name,
                        size // 1024 // 1024 // 1024,
                    )

                target.filesystems.add(CellebriteFilesystem(dump.path))


class CellebriteFilesystem(LayerFilesystem):
    """Cellebrite ``FileDump`` filesystem implementation."""

    __type__ = "cellebrite"

    def __init__(self, path: Path, base: str | None = None, **kwargs):
        super().__init__(**kwargs)
        self.source = path

        if path.suffix == ".zip":
            fs = ZipFilesystem(path.open("rb"), base=base)
        else:
            raise ValueError(f"Unsupported Cellebrite dump type {path.name}")

        # We add the full file system from ``/filesystem1`` as a layer to the root ``/`` and
        # mount found extras and extraction metadata, such as keychain dumps to folders in ``/$fs$/fs0``,
        # to keep this information accessible for device specific plugins.
        for fs_dir, dest in [
            ("/filesystem1", "/"),
            ("/extra", "/$fs$/fs0/extra"),
            ("/metadata1", "/$fs$/fs0/metadata"),
        ]:
            if fs.path(fs_dir).exists():
                # Mounts the ZipFilesystem at the provided fs_dir base. This way we can
                # (ab)use a single zip file handle for multiple filesystem layers.
                self.append_layer().mount(dest, fs, base=fs_dir)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.source}>"



from __future__ import annotations

import textwrap
from datetime import datetime, timezone
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.shell.wget import WgetPlugin
from dissect.target.plugins.os.unix._os import UnixPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_wget_hsts(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if .wget-hsts files are parsed as expected."""
    fs_unix.map_file_fh("/etc/hostname", BytesIO(b"example.domain"))

    fs_unix.map_file_fh(
        "/home/user/.wget-hsts",
        BytesIO(
            textwrap.dedent(
                """
        # HSTS 1.0 Known Hosts database for GNU Wget.
        # Edit at your own risk.
        # <hostname>	<port>	<incl. subdomains>	<created>	<max-age>
        mozilla.org	0	0	1717171717	31536000
        google.com	0	1	1711711711	31536000
        www.python.org	0	1	1713143141	63072000
        github.com	0	1	1713371337	31536000
        """
            ).encode()
        ),
    )

    target_unix_users.add_plugin(UnixPlugin)
    target_unix_users.add_plugin(WgetPlugin)

    results = sorted(target_unix_users.wget.hsts(), key=lambda r: r.host)

    assert len(results) == 4
    assert [r.host for r in results] == [
        "github.com",
        "google.com",
        "mozilla.org",
        "www.python.org",
    ]

    assert results[0].hostname == "example"
    assert results[0].domain == "domain"
    assert results[0].username == "user"
    assert results[0].host == "github.com"
    assert not results[0].explicit_port
    assert results[0].include_subdomains
    assert results[0].ts_created == datetime(2024, 4, 17, 16, 28, 57, tzinfo=timezone.utc)
    assert results[0].max_age == datetime(2025, 4, 17, 16, 28, 57, tzinfo=timezone.utc)
    assert results[0].source == "/home/user/.wget-hsts"



from __future__ import annotations

import plistlib
from typing import TYPE_CHECKING

from dissect.target import container
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class UtmLoader(Loader):
    """Load UTM virtual machine files."""

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        config_path = self.absolute_path.joinpath("config.plist")
        self.config: dict = plistlib.loads(config_path.read_bytes())

    @staticmethod
    def detect(path: Path) -> bool:
        return path.is_dir() and path.suffix.lower() == ".utm" and path.joinpath("config.plist").is_file()

    def map(self, target: Target) -> None:
        data_dir = self.absolute_path.joinpath("Data")
        for drive in self.config.get("Drive", []):
            if drive.get("ImageType") != "Disk":
                continue

            path = data_dir.joinpath(drive["ImageName"])
            try:
                target.disks.add(container.open(path))
            except Exception:
                target.log.exception("Failed to load drive: %s", drive)



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.ffs import c_ffs, ffs

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class FfsFilesystem(Filesystem):
    __type__ = "ffs"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.ffs = ffs.FFS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        for sb_offset in ffs.SBLOCKSEARCH:
            fh.seek(sb_offset)
            block = fh.read(4096)
            magic = int.from_bytes(block[1372:1376], "little")

            if magic in (c_ffs.c_ffs.FS_UFS1_MAGIC, c_ffs.c_ffs.FS_UFS2_MAGIC):
                return True

        return False

    def get(self, path: str) -> FilesystemEntry:
        return FfsFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: ffs.INode | None = None) -> ffs.INode:
        try:
            return self.ffs.get(path, node)
        except ffs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except ffs.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except ffs.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except ffs.Error as e:
            raise FileNotFoundError(path) from e


class FfsFilesystemEntry(FilesystemEntry):
    def get(self, path: str) -> FilesystemEntry:
        entry_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        entry = self.fs._get_node(path, self.entry)
        return FfsFilesystemEntry(self.fs, entry_path, entry)

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def _iterdir(self) -> Iterator[ffs.INode]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        if self.is_symlink():
            for entry in self.readlink_ext().iterdir():
                yield entry
        else:
            for entry in self.entry.iterdir():
                if entry.name in (".", ".."):
                    continue

                yield entry

    def iterdir(self) -> Iterator[str]:
        for entry in self._iterdir():
            yield entry.name

    def scandir(self) -> Iterator[FilesystemEntry]:
        for entry in self._iterdir():
            entry_path = fsutil.join(self.path, entry.name, alt_separator=self.fs.alt_separator)
            yield FfsFilesystemEntry(self.fs, entry_path, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_dir()
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.is_file()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        node = self.entry.inode

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = st_info = fsutil.stat_result(
            [
                self.entry.mode,
                self.entry.inum,
                id(self.fs),
                node.di_nlink,
                node.di_uid,
                node.di_gid,
                node.di_size,
                # timestamp() returns a float which will fill both the integer and float fields
                self.entry.atime.timestamp(),
                self.entry.mtime.timestamp(),
                self.entry.ctime.timestamp(),
            ]
        )

        # Note: stat on linux always returns the default block size of 4096
        # We are returning the actual block size of the filesystem, as on BSD
        st_info.st_blksize = self.fs.ffs.block_size
        # Note: st_blocks * 512 can be lower than st_blksize because FFS employs fragments
        st_info.st_blocks = self.entry.nblocks

        # Set the nanosecond resolution separately
        st_info.st_atime_ns = self.entry.atime_ns
        st_info.st_mtime_ns = self.entry.mtime_ns
        st_info.st_ctime_ns = self.entry.ctime_ns

        # FFS2 has a birth time, FFS1 does not
        if btime := self.entry.btime:
            st_info.st_birthtime = btime.timestamp()
            st_info.st_birthtime_ns = self.entry.btime_ns

        return st_info



from __future__ import annotations

import bz2
import datetime
import enum
import functools
import gzip
import json
from contextlib import contextmanager
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, Callable, TextIO

import structlog

try:
    import lz4.frame

    HAS_LZ4 = True
except ImportError:
    HAS_LZ4 = False


try:
    import zstandard

    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False


from flow.record import Record, RecordDescriptor, RecordStreamWriter
from flow.record.adapter.jsonfile import JsonfileWriter
from flow.record.jsonpacker import JsonRecordPacker

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.plugin import FunctionDescriptor
    from dissect.target.target import Target
    from dissect.target.tools.dump.run import RecordStreamElement
    from dissect.target.tools.dump.state import DumpState

log = structlog.get_logger(__name__)


class Compression(str, enum.Enum):
    """Supported compression types."""

    BZIP2 = "bzip2"
    GZIP = "gzip"
    LZ4 = "lz4"
    ZSTD = "zstandard"
    NONE = None


class Serialization(str, enum.Enum):
    """Supported serialization methods."""

    JSONLINES = "jsonlines"
    MSGPACK = "msgpack"


COMPRESSION_TO_EXT = {
    Compression.BZIP2: "bz2",
    Compression.GZIP: "gz",
    Compression.LZ4: "lz4",
    Compression.ZSTD: "zstd",
    Compression.NONE: "",
}


DEST_DIR_CACHE_SIZE = 10
DEST_FILENAME_CACHE_SIZE = 10
OPEN_WRITERS_LIMIT = 10


def get_nested_attr(obj: Any, nested_attr: str) -> Any:
    parts = nested_attr.split(".")
    return functools.reduce(getattr, [obj, *parts])


@lru_cache(maxsize=DEST_DIR_CACHE_SIZE)
def get_sink_dir_by_target(target: Target, function: FunctionDescriptor) -> Path:
    func_first_name, _, _ = function.name.partition(".")
    return Path(target.name) / func_first_name


@functools.lru_cache(maxsize=DEST_DIR_CACHE_SIZE)
def get_sink_dir_by_func(target: Target, function: FunctionDescriptor) -> Path:
    func_first_name, _, _ = function.name.partition(".")
    return Path(func_first_name) / target.name


def slugify_descriptor_name(descriptor_name: str) -> str:
    return descriptor_name.replace("/", "_")


@functools.lru_cache(maxsize=DEST_FILENAME_CACHE_SIZE)
def get_sink_filename(
    record_descriptor: RecordDescriptor,
    serialization: Serialization,
    compression: Compression | None = None,
) -> str:
    """Return a sink filename for provided record descriptor, serialization and compression."""
    record_type = slugify_descriptor_name(record_descriptor.name)

    serialization_details = SERIALIZERS[serialization]
    serialization_ext = serialization_details["ext"]

    parts = [record_type, serialization_ext]

    compression_ext = COMPRESSION_TO_EXT[compression]
    if compression_ext:
        parts.append(compression_ext)

    return ".".join(parts)


def get_relative_sink_path(
    element: RecordStreamElement, serialization: str, compression: Compression | None = None
) -> Path:
    """Return a sink path relative to an output directory."""
    sink_dir = get_sink_dir_by_target(element.target, element.func)
    sink_filename = get_sink_filename(element.record._desc, serialization, compression)
    return sink_dir / sink_filename


def open_path(path: Path, mode: str, compression: Compression | None = None) -> BinaryIO:
    """Open ``path`` using ``mode``, with specified ``compression`` and return a file object."""
    if compression == Compression.GZIP:
        fh = gzip.open(path, mode)  # noqa: SIM115
    elif compression == Compression.BZIP2:
        fh = bz2.open(path, mode)  # noqa: SIM115
    elif compression == Compression.LZ4:
        if not HAS_LZ4:
            raise ValueError("Python module lz4 is not available")
        fh = lz4.frame.open(path, mode)
    elif compression == Compression.ZSTD:
        if not HAS_ZSTD:
            raise ValueError("Python module zstandard is not available")
        cctx = zstandard.ZstdCompressor()
        fh = cctx.stream_writer(path.open(mode))
    elif compression == Compression.NONE:
        fh = path.open(mode)
    else:
        raise ValueError(f"Unrecognised compression method: {compression}")
    return fh


class JsonLinesWriter(JsonfileWriter):
    def __init__(self, fp: TextIO, **kwargs):
        self.fp = fp
        self.packer = SortedKeysJsonRecordPacker(indent=None)
        self.packer.on_descriptor.add_handler(self.packer_on_new_descriptor)

    def _write(self, obj: Record | RecordDescriptor) -> None:
        record_json = self.packer.pack(obj)
        line = record_json + "\n"
        line = line.encode("utf-8")
        self.fp.write(line)

    def flush(self) -> None:
        if not self.fp.closed:
            self.fp.flush()

    def close(self) -> None:
        self.fp.close()


class SortedKeysJsonRecordPacker(JsonRecordPacker):
    def pack(self, obj: Record | RecordDescriptor) -> str:
        return json.dumps(
            obj,
            default=self.pack_obj,
            indent=self.indent,
            sort_keys=True,
        )


SERIALIZERS = {
    Serialization.JSONLINES: {
        "writer": JsonLinesWriter,
        "ext": "jsonl",
    },
    Serialization.MSGPACK: {
        "writer": RecordStreamWriter,
        "ext": "rec",
    },
}


def get_sink_writer(
    full_sink_path: Path,
    serialization: Serialization,
    compression: Compression | None = None,
    new_sink: bool = True,
) -> JsonfileWriter | RecordStreamWriter:
    # create parent directories if they are missing
    full_sink_path.parent.mkdir(parents=True, exist_ok=True)

    mode = "wb" if new_sink else "ab"

    fh = open_path(full_sink_path, mode=mode, compression=compression)
    writer_cls = SERIALIZERS[serialization]["writer"]
    return writer_cls(fh)


@contextmanager
def cached_sink_writers(state: DumpState) -> Iterator[Callable]:
    # Poor man's cache that cleans up when it hits the limit of `OPEN_WRITERS_LIMIT`.
    # The cache is needed to reduce file handler open/close flickering for unsorted records stream.
    writers_cache = {}

    def close_writers() -> None:
        for path, writer in writers_cache.items():
            writer.close()
            log.debug("Sink writer closed", writer=writer, path=path)
        writers_cache.clear()

    def write_element(element: RecordStreamElement) -> int:
        sink_path = get_relative_sink_path(
            element,
            state.serialization,
            compression=state.compression,
        )
        element.sink_path = sink_path
        sink = state.get_sink(sink_path)

        new_sink = False
        if sink and not sink.is_dirty:
            log.error("Can not write to a clean finished sink", sink=sink, state=state.path)
            raise ValueError("Can not write to a clean finished sink")
        if not sink:
            sink = state.create_sink(sink_path, element)
            new_sink = True

        if sink.path in writers_cache:
            writer = writers_cache[sink.path]
        else:
            writer = writers_cache[sink.path] = get_sink_writer(
                state.get_full_sink_path(sink),
                state.serialization,
                compression=state.compression,
                new_sink=new_sink,
            )
            log.debug("Sink writer opened", writer=writer, path=sink.path)

        writer.write(element.record)
        fh_position = writer.fp.tell()

        state.update(element, fh_position)

        log.debug(
            "Record written to sink",
            sink_path=sink.path,
            record=element.record,
            writer=writer.__class__.__name__,
            position_bytes=fh_position,
            new_sink=new_sink,
        )

        if len(writers_cache) >= OPEN_WRITERS_LIMIT:
            close_writers()

        return fh_position

    try:
        yield write_element
    finally:
        close_writers()


def get_current_utc_time() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


def parse_datetime_iso(datetime_str: str) -> datetime.datetime:
    return datetime.datetime.fromisoformat(datetime_str)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target import container
from dissect.target.exceptions import TargetPathNotFoundError
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class RawLoader(Loader):
    """Load raw container files such as disk images."""

    def __init__(self, path: Path, **kwargs):
        if not path.exists():
            raise TargetPathNotFoundError("Provided target path does not exist")

        super().__init__(path, **kwargs)

    @staticmethod
    def detect(path: Path) -> bool:
        return not path.is_dir()

    def map(self, target: Target) -> None:
        target.disks.add(container.open(self.path))



from __future__ import annotations

import stat
from datetime import datetime, timezone
from enum import IntEnum
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.util import ts

from dissect.target.exceptions import FileNotFoundError, NotADirectoryError
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

try:
    from cbc_sdk.live_response_api import LiveResponseError, LiveResponseSession

    HAS_CARBON_BLACK = True
except ImportError:
    HAS_CARBON_BLACK = False

CB_TIMEFORMAT = "%Y-%m-%dT%H:%M:%S%fZ"


class OS(IntEnum):
    WINDOWS = 1
    LINUX = 2
    MAC = 4


class CbFilesystem(Filesystem):
    __type__ = "cb"

    def __init__(self, session: LiveResponseSession, prefix: str, *args, **kwargs):
        if not HAS_CARBON_BLACK:
            raise ImportError(
                "Required dependency 'carbon-black-cloud-sdk-python' is missing, install with 'pip install dissect.target[cb]'"  # noqa: E501
            )

        self.session = session
        self.prefix = prefix.lower()

        if self.session.os_type == OS.WINDOWS:
            alt_separator = "\\"
            case_sensitive = False
        else:
            alt_separator = ""
            case_sensitive = True

        super().__init__(*args, alt_separator=alt_separator, case_sensitive=case_sensitive, **kwargs)

    @staticmethod
    def detect(fh: BinaryIO) -> bool:
        raise TypeError("Detect is not allowed on CbFilesystem class")

    def get(self, path: str) -> CbFilesystemEntry:
        """Returns a CbFilesystemEntry object corresponding to the given path."""
        cbpath = fsutil.normalize(path, alt_separator=self.alt_separator).strip("/")
        if self.session.os_type == OS.WINDOWS:
            cbpath = cbpath.replace("/", "\\")

        cbpath = self.prefix + cbpath

        try:
            if cbpath == self.prefix:
                # Root entries behave funky, so make up our own entry
                entry = {
                    "filename": self.prefix,
                    "attributes": ["DIRECTORY"],
                    "last_access_time": "1970-01-01T00:00:00Z",
                    "last_write_time": "1970-01-01T00:00:00Z",
                    "create_time": "1970-01-01T00:00:00Z",
                    "size": 0,
                }
            else:
                res = self.session.list_directory(cbpath)
                if len(res) != 1:
                    raise FileNotFoundError(path)

                entry = res[0]

            return CbFilesystemEntry(self, path, entry, cbpath)
        except LiveResponseError:
            raise FileNotFoundError(path)


class CbFilesystemEntry(FilesystemEntry):
    def __init__(self, fs: Filesystem, path: str, entry: Any, cbpath: str):
        super().__init__(fs, path, entry)
        self.cbpath = cbpath

    def get(self, path: str) -> CbFilesystemEntry:
        """Get a filesystem entry relative from the current one."""
        full_path = fsutil.join(self.path, path)
        return self.fs.get(full_path)

    def open(self) -> BinaryIO:
        """Returns file handle (file-like object)."""
        return self.fs.session.get_raw_file(self.cbpath)

    def iterdir(self) -> Iterator[str]:
        """List the directory contents of a directory. Returns a generator of strings."""
        for entry in self.scandir():
            yield entry.name

    def scandir(self) -> Iterator[CbFilesystemEntry]:
        """List the directory contents of this directory. Returns a generator of filesystem entries."""
        if not self.is_dir():
            raise NotADirectoryError(f"'{self.path}' is not a directory")

        separator = self.fs.alt_separator or "/"
        for entry in self.fs.session.list_directory(self.cbpath + separator):
            filename = entry["filename"]
            if filename in (".", ".."):
                continue

            path = fsutil.join(self.path, filename, alt_separator=self.fs.alt_separator)
            cbpath = separator.join([self.cbpath, filename])
            yield CbFilesystemEntry(self.fs, path, entry, cbpath)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        return "DIRECTORY" in self.entry["attributes"]

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        return not self.is_dir()

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return False

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        mode = stat.S_IFDIR if self.is_dir() else stat.S_IFREG

        atime = ts.to_unix(_parse_ts(self.entry["last_access_time"]))
        mtime = ts.to_unix(_parse_ts(self.entry["last_write_time"]))
        ctime = ts.to_unix(_parse_ts(self.entry["create_time"]))

        # ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
        st_info = [
            mode | 0o755,
            fsutil.generate_addr(self.cbpath),
            id(self.fs),
            1,
            0,
            0,
            self.entry["size"],
            atime,
            mtime,
            ctime,
        ]
        return fsutil.stat_result(st_info)


def _parse_ts(ts: str) -> datetime:
    return datetime.strptime(ts, CB_TIMEFORMAT).replace(tzinfo=timezone.utc)



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.helpers.record import (
    MacOSInterfaceRecord,
    UnixInterfaceRecord,
    WindowsInterfaceRecord,
)
from dissect.target.plugins.os.default.network import InterfaceRecord, NetworkPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.fixture(params=[MacOSInterfaceRecord, WindowsInterfaceRecord, UnixInterfaceRecord])
def network_record(request: pytest.FixtureRequest) -> InterfaceRecord:
    return request.param(
        name="interface_name",
        type="physical",
        enabled=True,
        cidr=["10.42.42.10/24"],
        gateway=["10.42.42.1"],
        dns=["8.8.8.8", "1.1.1.1"],
        mac=["DE:AD:BE:EF:00:00"],
        source="some_file",
    )


def test_base_network_plugin(target_bare: Target, network_record: InterfaceRecord) -> None:
    with patch.object(NetworkPlugin, "_interfaces", return_value=[network_record]):
        network = NetworkPlugin(target_bare)
        interfaces = list(network.interfaces())
        assert len(interfaces) == 1

        assert network.ips() == ["10.42.42.10"]
        assert network.gateways() == ["10.42.42.1"]
        assert network.macs() == ["DE:AD:BE:EF:00:00"]
        assert sorted(map(str, network.dns())) == ["1.1.1.1", "8.8.8.8"]

        assert len(list(network.in_cidr("10.42.42.0/24"))) == 1
        assert len(list(network.in_cidr("10.43.42.0/24"))) == 0

        assert len(list(network.with_mac("DE:AD:BE:EF:00:00"))) == 1
        assert len(list(network.with_mac("DE:AD:BE:EF:00:01"))) == 0

        assert len(list(network.with_ip("10.42.42.10"))) == 1
        assert len(list(network.with_ip("10.42.42.42"))) == 0



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.hypervisor import vmdk

from dissect.target.container import Container

if TYPE_CHECKING:
    from pathlib import Path


class VmdkContainer(Container):
    """VMware virtual hard disks."""

    __type__ = "vmdk"

    def __init__(self, fh: BinaryIO | Path, *args, **kwargs):
        self.vmdk = vmdk.VMDK(fh)
        super().__init__(fh, self.vmdk.size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return fh.read(4) in (b"KDMV", b"COWD", b"# Di")

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return path.suffix.lower().endswith(".vmdk")

    def read(self, length: int) -> bytes:
        return self.vmdk.read(length)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.vmdk.seek(offset, whence)

    def tell(self) -> int:
        return self.vmdk.tell()

    def close(self) -> None:
        pass



from __future__ import annotations

from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import NamespacePlugin

GENERIC_LOG_RECORD_FIELDS = [
    ("datetime", "ts"),
    ("string", "message"),
    ("path", "source"),
]

GENERIC_FILE_TRANSFER_RECORD_FIELDS = [
    ("datetime", "ts"),
    ("path", "filename"),
    ("string", "message"),
    ("path", "source"),
]

RemoteAccessLogRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "remoteaccess/log", GENERIC_LOG_RECORD_FIELDS
)

RemoteAccessFileTransferRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "remoteaccess/filetransfer", GENERIC_FILE_TRANSFER_RECORD_FIELDS
)


class RemoteAccessPlugin(NamespacePlugin):
    """General Remote Access plugin.

    This plugin groups the functions of all remote access plugins. For example,
    instead of having to run both teamviewer.remoteaccess and anydesk.remoteaccess,
    you only have to run remoteaccess.remoteaccess to get output from both tools.
    """

    __namespace__ = "remoteaccess"



from __future__ import annotations

import pytest

from dissect.target.helpers.network import IANAProtocol


@pytest.mark.parametrize(
    ("input_number", "expected_output"),
    [
        (0, "HOPOPT"),
        (6, "TCP"),
        (17, "UDP"),
        (27, "RDP"),
        (148, "UNASSIGNED_148"),
        (169, "UNASSIGNED_169"),
        (252, "UNASSIGNED_252"),
        (253, "USE_FOR_EXPERIMENTATION_AND_TESTING_253"),
        (255, "RESERVED_255"),
        (1337, "UNKNOWN_1337"),
    ],
)
def test_helpers_network_iana_protocol_translation(input_number: int, expected_output: str) -> None:
    """Test if we translate IANA protocol numbers correctly."""

    assert IANAProtocol(input_number).name == expected_output
    assert IANAProtocol(input_number).value == input_number



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target import Target


def find_containers(paths: list[Path]) -> Iterator[Path]:
    for path in paths:
        for config_path in path.iterdir():
            if (config_file := config_path.joinpath("colima.yaml")).exists():
                name = f"-{config_file.parts[-2]}" if config_file.parts[-2] != "default" else ""
                if (disk_path := path.joinpath("_lima", f"colima{name}", "diffdisk")).exists():
                    yield disk_path


class ColimaChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields Colima containers.

    Colima is a container runtime for macOS and Linux.

    References:
        - https://github.com/abiosoft/colima/blob/5d2e91c4a491d4ae35d69fb2583f4f959401bc37
    """

    __type__ = "colima"

    def __init__(self, target: Target):
        super().__init__(target)
        self.paths = [
            path
            for user in self.target.user_details.all_with_home()
            if (path := user.home_path.joinpath(".colima")).exists()
        ]

    def check_compatible(self) -> None:
        if not self.paths:
            raise UnsupportedPluginError("No Colima configurations found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for container in find_containers(self.paths):
            yield ChildTargetRecord(
                type=self.__type__,
                path=container,
                _target=self.target,
            )



from __future__ import annotations

from typing import TYPE_CHECKING

from flow.record.fieldtypes import posix_path

from dissect.target.plugin import OperatingSystem
from dissect.target.plugins.os.unix.bsd.darwin.macos._os import MacOSPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_unix_bsd_darwin_macos_os(target_macos_users: Target, fs_macos: VirtualFilesystem) -> None:
    """Test if we detect a macOS target correctly."""

    target_macos_users.add_plugin(MacOSPlugin)

    interface = absolute_path("_data/plugins/os/unix/bsd/darwin/macos/_os/en0.plist")
    fs_macos.map_file("/private/var/db/dhcpclient/leases/en0.plist", interface)

    users = list(target_macos_users.users())
    ips = target_macos_users.ips
    ips.sort()

    dissect_user = users[0]
    test_user = users[1]

    assert target_macos_users.os == OperatingSystem.MACOS
    assert target_macos_users.hostname == "dummys Mac"
    assert target_macos_users.version == "macOS 11.7.5 (20G1225)"

    assert len(users) == 2
    assert len(ips) == 2

    assert dissect_user._desc.name == "macos/user"
    assert dissect_user.name == "_dissect"
    assert dissect_user.passwd == "*"
    assert dissect_user.home == posix_path("/Users/dissect")
    assert dissect_user.shell == "/usr/bin/false"
    assert dissect_user.source == "/var/db/dslocal/nodes/Default/users/_dissect.plist"

    assert test_user.home is None

    assert ips == ["10.42.43.63", "10.42.43.64"]



from __future__ import annotations

import shutil
from typing import TYPE_CHECKING, Callable

import pytest

from dissect.target.filesystems.ntfs import NtfsFilesystem
from dissect.target.loader import open as loader_open
from dissect.target.loaders.velociraptor import VelociraptorLoader
from dissect.target.target import Target
from tests._utils import absolute_path, mkdirs

if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture(params=["mft", "ntfs", "ntfs_vss", "lazy_ntfs", "auto"])
def mock_velociraptor_dir(request: pytest.FixtureRequest, tmp_path: Path) -> Path:
    sub_dir = request.param
    paths = [
        f"uploads/{sub_dir}/%5C%5C.%5CC%3A/",
        f"uploads/{sub_dir}/%5C%5C.%5CC%3A/$Extend",
        f"uploads/{sub_dir}/%5C%5C.%5CC%3A/windows/system32",
        f"uploads/{sub_dir}/%5C%5C%3F%5CGLOBALROOT%5CDevice%5CHarddiskVolumeShadowCopy1/",
        f"uploads/{sub_dir}/%5C%5C%3F%5CGLOBALROOT%5CDevice%5CHarddiskVolumeShadowCopy1/$Extend",
        f"uploads/{sub_dir}/%5C%5C%3F%5CGLOBALROOT%5CDevice%5CHarddiskVolumeShadowCopy1/windows/system32",
        f"uploads/{sub_dir}/%5C%5C.%5CC%3A/%2ETEST",
        "results",
    ]
    root = tmp_path
    mkdirs(root, paths)

    (root / "uploads.json").write_bytes(b"{}")
    (root / f"uploads/{sub_dir}/%5C%5C.%5CC%3A/C-DRIVE.txt").write_bytes(b"{}")
    (root / f"uploads/{sub_dir}/%5C%5C.%5CC%3A/Microsoft-Windows-Windows Defender%254WHC.evtx").write_bytes(b"{}")
    (root / f"uploads/{sub_dir}/%5C%5C.%5CC%3A/other.txt").write_text("my first file")

    with absolute_path("_data/plugins/filesystem/ntfs/mft/mft.raw").open("rb") as fh:
        mft = fh.read(10 * 1024)

    root.joinpath(paths[0]).joinpath("$MFT").write_bytes(mft)
    root.joinpath(paths[3]).joinpath("$MFT").write_bytes(mft)

    # Add one record so we can test if it works
    data = bytes.fromhex(
        "5800000002000000c100000000000100bf000000000001002003010000000000"
        "6252641a86a4d7010381008000000000000000002000000018003c0069007300"
        "2d00310035005000320036002e0074006d00700000000000"
    )
    root.joinpath(paths[1]).joinpath("$UsnJrnl%3A$J").write_bytes(data)
    root.joinpath(paths[4]).joinpath("$UsnJrnl%3A$J").write_bytes(data)

    return root


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_velociraptor_dir: Path) -> None:
    """Test that we correctly use ``VelociraptorLoader`` when opening a ``Target``."""
    path = mock_velociraptor_dir

    target = opener(path)
    assert isinstance(target._loader, VelociraptorLoader)
    assert target.path == path


def test_windows_ntfs(mock_velociraptor_dir: Path) -> None:
    """Test that ``VelociraptorLoader`` correctly loads a Windows directory structure."""
    loader = loader_open(mock_velociraptor_dir)
    assert isinstance(loader, VelociraptorLoader)

    t = Target()
    loader.map(t)
    t.apply()

    assert "sysvol" in t.fs.mounts
    assert "c:" in t.fs.mounts

    usnjrnl_records = 0
    for fs in t.filesystems:
        if isinstance(fs, NtfsFilesystem):
            usnjrnl_records += len(list(fs.ntfs.usnjrnl.records()))
    assert usnjrnl_records == 2
    assert len(t.filesystems) == 4

    assert t.fs.path("sysvol/C-DRIVE.txt").exists()
    assert t.fs.path("sysvol/other.txt").read_text() == "my first file"
    assert t.fs.path("sysvol/.TEST").exists()
    assert t.fs.path("sysvol/Microsoft-Windows-Windows Defender%254WHC.evtx").exists()


def test_windows_ntfs_zip(mock_velociraptor_dir: Path) -> None:
    """Test that ``VelociraptorLoader`` correctly loads a Windows ZIP structure."""
    shutil.make_archive(mock_velociraptor_dir.joinpath("test_ntfs"), "zip", mock_velociraptor_dir)

    path = mock_velociraptor_dir.joinpath("test_ntfs.zip")
    loader = loader_open(path)
    assert isinstance(loader, VelociraptorLoader)

    t = Target()
    loader.map(t)
    t.apply()

    assert "sysvol" in t.fs.mounts
    assert "c:" in t.fs.mounts

    usnjrnl_records = 0
    for fs in t.filesystems:
        if isinstance(fs, NtfsFilesystem):
            usnjrnl_records += len(list(fs.ntfs.usnjrnl.records()))
    assert usnjrnl_records == 2
    assert len(t.filesystems) == 4
    assert t.fs.path("sysvol/C-DRIVE.txt").exists()
    assert t.fs.path("sysvol/.TEST").exists()
    assert t.fs.path("sysvol/Microsoft-Windows-Windows Defender%4WHC.evtx").exists()


@pytest.mark.parametrize(
    "paths",
    [
        (["uploads/file/etc", "uploads/file/var", "uploads/file/%2ETEST"]),
        (["uploads/auto/etc", "uploads/auto/var", "uploads/auto/%2ETEST"]),
        (["uploads/file/etc", "uploads/file/var", "uploads/file/opt", "uploads/file/%2ETEST"]),
        (["uploads/auto/etc", "uploads/auto/var", "uploads/auto/opt", "uploads/auto/%2ETEST"]),
        (["uploads/file/Library", "uploads/file/Applications", "uploads/file/%2ETEST"]),
        (["uploads/auto/Library", "uploads/auto/Applications", "uploads/auto/%2ETEST"]),
    ],
)
def test_unix(paths: list[str], tmp_path: Path) -> None:
    """Test that ``VelociraptorLoader`` correctly loads a Unix directory structure."""
    root = tmp_path
    mkdirs(root, paths)

    (root / "uploads.json").write_bytes(b"{}")

    loader = loader_open(root)
    assert isinstance(loader, VelociraptorLoader)

    t = Target()
    loader.map(t)
    t.apply()

    assert len(t.filesystems) == 1
    assert t.fs.path("/.TEST").exists()


@pytest.mark.parametrize(
    "paths",
    [
        (["uploads/file/etc", "uploads/file/var", "uploads/file/%2ETEST"]),
        (["uploads/auto/etc", "uploads/auto/var", "uploads/auto/%2ETEST"]),
        (["uploads/file/etc", "uploads/file/var", "uploads/file/opt", "uploads/file/%2ETEST"]),
        (["uploads/auto/etc", "uploads/auto/var", "uploads/auto/opt", "uploads/auto/%2ETEST"]),
        (["uploads/file/Library", "uploads/file/Applications", "uploads/file/%2ETEST"]),
        (["uploads/auto/Library", "uploads/auto/Applications", "uploads/auto/%2ETEST"]),
    ],
)
def test_unix_zip(paths: list[str], tmp_path: Path) -> None:
    """Test that ``VelociraptorLoader`` correctly loads a Unix ZIP structure."""
    root = tmp_path
    mkdirs(root, paths)

    (root / "uploads.json").write_bytes(b"{}")

    shutil.make_archive(tmp_path.joinpath("test_unix"), "zip", tmp_path)

    path = tmp_path.joinpath("test_unix.zip")
    loader = loader_open(path)
    assert isinstance(loader, VelociraptorLoader)

    t = Target()
    loader.map(t)
    t.apply()

    assert len(t.filesystems) == 1
    assert t.fs.path("/.TEST").exists()



project = "dissect.target"

extensions = [
    "autoapi.extension",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.doctest",
    "sphinx.ext.napoleon",
    "sphinx_argparse_cli",
]

exclude_patterns = []

html_theme = "furo"

autoapi_type = "python"
autoapi_dirs = ["../../dissect/"]
autoapi_ignore = ["*tests*", "*.tox*", "*venv*", "*examples*"]
autoapi_python_use_implicit_namespaces = True
autoapi_add_toctree_entry = False
autoapi_root = "api"
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    "imported-members",
]
autoapi_keep_files = True
autoapi_template_dir = "_templates/autoapi"

autodoc_typehints = "signature"
autodoc_member_order = "groupwise"

autosectionlabel_prefix_document = True

suppress_warnings = [
    # https://github.com/readthedocs/sphinx-autoapi/issues/285
    "autoapi.python_import_resolution",
]



from __future__ import annotations

from dissect.target.helpers.record import TargetRecordDescriptor

BootTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/boot",
    [],
)

CalendarTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/calendar",
    [],
)

DailyTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/daily",
    [
        ("uint16", "days_between_triggers"),
        ("uint16[]", "unused"),
    ],
)

ComHandlerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/action/comhandler",
    [
        ("string", "action_type"),
        ("string", "class_id"),
        ("string", "com_data"),
    ],
)

EventTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/event",
    [
        ("string", "subscription"),
        ("string", "period_of_occurence"),
        ("uint16", "number_of_occurences"),
        ("string", "matching_elements"),
        ("string", "value_queries"),
    ],
)

ExecRecord = TargetRecordDescriptor(
    "filesystem/windows/task/action/exec",
    [
        ("string", "action_type"),
        ("string", "command"),
        ("string", "arguments"),
        ("string", "working_directory"),
    ],
)

IdleTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/idle",
    [],
)

LogonTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/logon",
    [
        ("string", "user_id"),
    ],
)

TimeTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/time",
    [],
)

BaseTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/base",
    [
        ("boolean", "trigger_enabled"),
        ("datetime", "start_boundary"),
        ("datetime", "end_boundary"),
        ("string", "repetition_interval"),
        ("string", "repetition_duration"),
        ("boolean", "repetition_stop_duration_end"),
        ("string", "execution_time_limit"),
        ("string", "delay"),
        ("string", "random_delay"),
        ("string", "trigger_data"),
    ],
)

MonthlyDateTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/monthly",
    [
        ("uint16[]", "day_of_month"),
        ("string[]", "months_of_year"),
    ],
)

MonthlyDowTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/monthly_dow",
    [
        ("uint16[]", "which_week"),
        ("string[]", "days_of_week"),
        ("string[]", "months_of_year"),
    ],
)

PaddingTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/padding",
    [
        ("uint16", "padding"),
        ("uint16", "reserved2"),
        ("uint16", "reserved3"),
    ],
)

SendEmailRecord = TargetRecordDescriptor(
    "filesystem/windows/task/action/send_email",
    [
        ("string", "action_type"),
        ("string", "server"),
        ("string", "subject"),
        ("string", "to"),
        ("string", "cc"),
        ("string", "bcc"),
        ("string", "replyto"),
        ("string", "email_from"),
        ("string", "header_name"),
        ("string", "header_value"),
        ("string", "body"),
        ("string", "attachment"),
    ],
)

SessionStateChangeTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/session_state_change",
    [
        ("string", "user_id"),
        ("string", "state_change"),
    ],
)

ShowMessageRecord = TargetRecordDescriptor(
    "filesystem/windows/task/action/show_message",
    [
        ("string", "tile"),
        ("string", "body"),
    ],
)

WeeklyTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/weekly",
    [
        ("uint16", "weeks_between_triggers"),
        ("string[]", "days_of_week"),
        ("uint16[]", "unused"),
    ],
)

WnfTriggerRecord = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/wnf",
    [
        ("string", "state_name"),
    ],
)

RegistrationTrigger = TargetRecordDescriptor(
    "filesystem/windows/task/trigger/registration",
    [
        ("datetime", "date"),
    ],
)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.kape import KapeLoader
from dissect.target.target import Target
from tests._utils import absolute_path, mkdirs

if TYPE_CHECKING:
    from pathlib import Path


@pytest.fixture
def mock_kape_dir(tmp_path: Path) -> Path:
    root = tmp_path
    mkdirs(root, ["C/windows/system32", "C/$Extend", "D/test", "E/test"])

    # Only need this to exist up until the root directory record to make dissect.ntfs happy
    with absolute_path("_data/plugins/filesystem/ntfs/mft/mft.raw").open("rb") as fh:
        (root / "C/$MFT").write_bytes(fh.read(10 * 1024))

    # Add one record so we can test if it works
    data = bytes.fromhex(
        "5800000002000000c100000000000100bf000000000001002003010000000000"
        "6252641a86a4d7010381008000000000000000002000000018003c0069007300"
        "2d00310035005000320036002e0074006d00700000000000"
    )
    (root / "C/$Extend/$J").write_bytes(data)

    return root


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target], mock_kape_dir: Path) -> None:
    """Test that we correctly use ``KapeLoader`` when opening a ``Target``."""
    for path in [mock_kape_dir, absolute_path("_data/loaders/kape/test.vhdx")]:
        target = opener(path)
        assert isinstance(target._loader, KapeLoader)
        assert target.path == path


@patch("dissect.target.filesystems.dir.DirectoryFilesystem.ntfs", None, create=True)
def test_dir(mock_kape_dir: Path) -> None:
    """Test the ``KapeLoader`` with a directory."""
    loader = loader_open(mock_kape_dir)
    assert isinstance(loader, KapeLoader)

    t = Target()
    loader.map(t)
    t.apply()

    assert "sysvol" in t.fs.mounts
    assert "c:" in t.fs.mounts
    assert "d:" in t.fs.mounts
    assert "e:" in t.fs.mounts

    # The 3 found drive letter directories + the fake NTFS filesystem
    assert len(t.filesystems) == 4
    # The 3 found drive letters + sysvol + the fake NTFS filesystem at /$fs$
    assert len(t.fs.mounts) == 5
    assert len(list(t.fs.mounts["c:"].ntfs.usnjrnl.records())) == 1


def test_vhdx() -> None:
    """Test the ``KapeLoader`` with a VHDX file."""
    path = absolute_path("_data/loaders/kape/test.vhdx")

    loader = loader_open(path)
    assert KapeLoader.detect(path)

    t = Target()
    loader.map(t)
    t.apply()

    assert "sysvol" in t.fs.mounts
    assert "c:" in t.fs.mounts



from __future__ import annotations

import base64
import itertools
import json
from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from dissect.sql import sqlite3
from dissect.sql.exceptions import Error as SQLError
from dissect.util.ts import webkittimestamp

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.fsutil import TargetPath, join
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.apps.browser.browser import (
    GENERIC_COOKIE_FIELDS,
    GENERIC_DOWNLOAD_RECORD_FIELDS,
    GENERIC_EXTENSION_RECORD_FIELDS,
    GENERIC_HISTORY_RECORD_FIELDS,
    GENERIC_PASSWORD_RECORD_FIELDS,
    BrowserPlugin,
    try_idna,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.sql.sqlite3 import SQLite3

    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

try:
    from Crypto.Cipher import AES, ChaCha20_Poly1305
    from Crypto.Protocol.KDF import PBKDF2
    from Crypto.Util.Padding import unpad

    HAS_CRYPTO = True

except ImportError:
    HAS_CRYPTO = False


CHROMIUM_DOWNLOAD_RECORD_FIELDS = [
    ("uri", "tab_url"),
    ("uri", "tab_referrer_url"),
    ("string", "mime_type"),
]

# References:
# - Reversing ``PostProcessData`` in ``elevation_service.exe``
# - https://chromium.googlesource.com/chromium/src/+/master/chrome/elevation_service/elevator.cc
elevation_def = """
struct Envelope {
    uint32  program_len;
    char    program[program_len];
    uint32  ciphertext_len;
    char    ciphertext[ciphertext_len]; // basically until EOF
};
struct GoogleChromeCipher {
    uint8   flag;                       // 0x01 = AES GCM, 0x02 = ChaCha20 Poly1305
    char    iv[12];
    char    ciphertext[32];
    char    mac_tag[16];
};
"""
c_elevation = cstruct(endian="<").load(elevation_def)

# References:
# - https://github.com/chromium/chromium/blob/main/components/download/public/common/download_item.h
DOWNLOAD_STATES = {
    0: "in_progress",
    1: "complete",
    2: "cancelled",
    3: "interrupted",
    4: "interrupted",  # Older versions of Chromium can have DownloadState value 4 as interrupted.
}


class ChromiumMixin:
    """Mixin class with methods for Chromium-based browsers."""

    DIRS = ()

    BrowserHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chromium/history", GENERIC_HISTORY_RECORD_FIELDS
    )

    BrowserCookieRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chromium/cookie", GENERIC_COOKIE_FIELDS
    )

    BrowserDownloadRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chromium/download", GENERIC_DOWNLOAD_RECORD_FIELDS + CHROMIUM_DOWNLOAD_RECORD_FIELDS
    )

    BrowserExtensionRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chromium/extension", GENERIC_EXTENSION_RECORD_FIELDS
    )

    BrowserPasswordRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chromium/password", GENERIC_PASSWORD_RECORD_FIELDS
    )

    def _build_userdirs(self, hist_paths: list[str]) -> list[tuple[UserDetails, TargetPath]]:
        """Join the selected browser dirs with the user home path.

        Args:
            hist_paths: A list with browser paths as strings.

        Returns:
            List of tuples containing user and file path objects.
        """
        users_dirs: list[tuple] = []
        for user_details in self.target.user_details.all_with_home():
            for d in hist_paths:
                home_dir: TargetPath = user_details.home_path
                for cur_dir in home_dir.glob(d):
                    cur_dir = cur_dir.resolve()
                    if not cur_dir.exists() or (user_details.user, cur_dir) in users_dirs:
                        continue
                    users_dirs.append((user_details, cur_dir))
        return users_dirs

    def _iter_db(
        self, filename: str, subdirs: list[str] | None = None
    ) -> Iterator[tuple[UserDetails, TargetPath, SQLite3]]:
        """Generate a connection to a sqlite database file.

        Args:
            filename: The filename as string of the database where the data is stored.
            subdirs: Subdirectories to also try for every configured directory.

        Yields:
            opened db_file (SQLite3)

        Raises:
            FileNotFoundError: If the history file could not be found.
            SQLError: If the history file could not be opened.
        """
        seen = set()
        dirs = list(self.DIRS)
        if subdirs:
            dirs.extend([join(dir, subdir) for dir, subdir in itertools.product(self.DIRS, subdirs)])

        for user, cur_dir in self._build_userdirs(dirs):
            db_file = cur_dir.joinpath(filename)

            if db_file in seen:
                continue
            seen.add(db_file)

            try:
                yield user, db_file, sqlite3.SQLite3(db_file.open())
            except FileNotFoundError:
                self.target.log.warning("Could not find %s file: %s", filename, db_file)
            except SQLError as e:
                self.target.log.warning("Could not open %s file: %s", filename, db_file)
                self.target.log.debug("", exc_info=e)

    def _iter_json(self, filename: str) -> Iterator[tuple[UserDetails, TargetPath, dict]]:
        """Iterate over all JSON files in the user directories, yielding a tuple
        of username, JSON file path, and the parsed JSON data.

        Args:
            filename (str): The name of the JSON file to search for in each
            user directory.

        Yields:
            A tuple containing the name of the user, the path to the JSON file, and the parsed JSON data.

        Raises:
            FileNotFoundError: If the json file could not be found.
        """
        for user, cur_dir in self._build_userdirs(self.DIRS):
            json_file = cur_dir.joinpath(filename)
            try:
                yield user, json_file, json.load(json_file.open())
            except FileNotFoundError:
                self.target.log.warning("Could not find %s file: %s", filename, json_file)

    def check_compatible(self) -> None:
        if not self._build_userdirs(self.DIRS):
            raise UnsupportedPluginError("No Chromium-based browser directories found")

    def history(self, browser_name: str | None = None) -> Iterator[BrowserHistoryRecord]:
        """Return browser history records from supported Chromium-based browsers.

        Args:
            browser_name: The name of the browser as a string.

        Yields:

        .. code-block:: text

            Records with the following fields:
                ts (datetime): Visit timestamp.
                browser (string): The browser from which the records are generated from.
                id (string): Record ID.
                url (uri): History URL.
                title (string): Page title.
                description (string): Page description.
                host (string): Hostname.
                visit_type (varint): Visit type.
                visit_count (varint): Amount of visits.
                hidden (string): Hidden value.
                typed (boolean): Typed value.
                session (varint): Session value.
                from_visit (varint): Record ID of the "from" visit.
                from_url (uri): URL of the "from" visit.
                source: (path): The source file of the history record.
        """
        for user, db_file, db in self._iter_db("History"):
            try:
                urls = {row.id: row for row in db.table("urls").rows()}
                visits = {}

                for row in db.table("visits").rows():
                    visits[row.id] = row
                    url = urls[row.url]

                    if row.from_visit and row.from_visit in visits:
                        from_visit = visits[row.from_visit]
                        from_url = urls[from_visit.url]
                    else:
                        from_visit, from_url = None, None

                    yield self.BrowserHistoryRecord(
                        ts=webkittimestamp(row.visit_time),
                        browser=browser_name,
                        id=row.id,
                        url=try_idna(url.url),
                        title=url.title,
                        description=None,
                        host=None,
                        visit_type=None,
                        visit_count=url.visit_count,
                        hidden=url.hidden,
                        typed=None,
                        session=None,
                        from_visit=row.from_visit or None,
                        from_url=try_idna(from_url.url) if from_url else None,
                        source=db_file,
                        _target=self.target,
                        _user=user.user,
                    )
            except SQLError as e:  # noqa: PERF203
                self.target.log.warning("Error processing history file: %s", db_file)
                self.target.log.debug("", exc_info=e)

    def cookies(self, browser_name: str | None = None) -> Iterator[BrowserCookieRecord]:
        """Return browser cookie records from supported Chromium-based browsers.

        Attempts to decrypt cookie values where possible.

        Args:
            browser_name: The name of the browser as a string.

        Yields:

        .. code-block:: text

            Records with the following fields:
                ts_created (datetime): Cookie created timestamp.
                ts_last_accessed (datetime): Cookie last accessed timestamp.
                browser (string): The browser from which the records are generated from.
                name (string): The cookie name.
                value (string): The cookie value.
                host (string): Cookie host key.
                path (string): Cookie path.
                expiry (varint): Cookie expiry.
                is_secure (bool): Cookie secury flag.
                is_http_only (bool): Cookie http only flag.
                same_site (bool): Cookie same site flag.
        """
        for user, db_file, db in self._iter_db("Cookies", subdirs=["Network"]):
            keys = {}

            if self.target.os == OperatingSystem.WINDOWS.value:
                try:
                    local_state_parent = db_file.parent.parent
                    if db_file.parent.name == "Network":
                        local_state_parent = local_state_parent.parent
                    local_state_path = local_state_parent.joinpath("Local State")

                    keys = self.decryption_keys(local_state_path, user.user.name)
                except ValueError as e:
                    self.target.log.warning("Failed to decrypt local state key in %s: %s", local_state_path, e)
                    self.target.log.debug("", exc_info=e)

            try:
                for cookie in db.table("cookies").rows():
                    cookie_value = cookie.value

                    if not cookie_value and keys and (encrypted_cookie := cookie.get("encrypted_value")):
                        try:
                            cookie_value = self.decrypt_value(user, keys, encrypted_cookie)
                        except (ValueError, UnicodeDecodeError) as e:
                            self.target.log.warning(
                                "Failed to decrypt cookie value for %s %s: %s", cookie.host_key, cookie.name, e
                            )

                        # Strip extra data
                        if cookie_value and encrypted_cookie[0:3] == b"v20":
                            cookie_value = cookie_value[32:]

                    yield self.BrowserCookieRecord(
                        ts_created=webkittimestamp(cookie.creation_utc),
                        ts_last_accessed=webkittimestamp(cookie.last_access_utc),
                        browser=browser_name,
                        name=cookie.name,
                        value=cookie_value,
                        host=cookie.host_key,
                        path=cookie.path,
                        expiry=int(cookie.has_expires),
                        is_secure=bool(cookie.is_secure),
                        is_http_only=bool(cookie.is_httponly),
                        same_site=bool(cookie.samesite),
                        source=db_file,
                        _target=self.target,
                        _user=user.user,
                    )
            except SQLError as e:
                self.target.log.warning("Error processing cookie file %s: %s", db_file, e)
                self.target.log.debug("", exc_info=e)

    def downloads(self, browser_name: str | None = None) -> Iterator[BrowserDownloadRecord]:
        """Return browser download records from supported Chromium-based browsers.

        Args:
            browser_name: The name of the browser as a string.

        Yields:

        .. code-block:: text

            Records with the following fields:
                ts_start (datetime): Download start timestamp.
                ts_end (datetime): Download end timestamp.
                browser (string): The browser from which the records are generated from.
                id (string): Record ID.
                path (string): Download path.
                url (uri): Download URL.
                tab_url (string): Tab URL.
                tab_referrer_url (string): Referrer URL.
                size (varint): Download file size.
                mime_type (string): MIME type.
                state (varint): Download state number.
                source: (path): The source file of the download record.
        """
        for user, db_file, db in self._iter_db("History"):
            try:
                download_chains = defaultdict(list)
                for row in db.table("downloads_url_chains"):
                    download_chains[row.id].append(row)

                for chain in download_chains.values():
                    chain.sort(key=lambda row: row.chain_index)

                for row in db.table("downloads").rows():
                    if download_path := row.target_path:
                        download_path = self.target.fs.path(download_path)

                    url = None
                    download_chain = download_chains.get(row.id)

                    if download_chain:
                        url = download_chain[-1].url
                        url = try_idna(url)

                    # https://github.com/chromium/chromium/blob/main/components/download/public/common/download_item.h
                    if state := row.get("state"):
                        state = DOWNLOAD_STATES.get(state)

                    yield self.BrowserDownloadRecord(
                        ts_start=webkittimestamp(row.start_time),
                        ts_end=webkittimestamp(row.end_time) if row.end_time else None,
                        browser=browser_name,
                        id=row.get("id"),
                        tab_url=try_idna(row.get("tab_url")),
                        tab_referrer_url=try_idna(row.get("tab_referrer_url")),
                        path=download_path,
                        url=url,
                        size=row.get("total_bytes"),
                        mime_type=row.get("mime_type"),
                        state=state,
                        source=db_file,
                        _target=self.target,
                        _user=user.user,
                    )
            except SQLError as e:  # noqa: PERF203
                self.target.log.warning("Error processing history file: %s", db_file)
                self.target.log.debug("", exc_info=e)

    def extensions(self, browser_name: str | None = None) -> Iterator[BrowserExtensionRecord]:
        """Iterates over all installed extensions for a given browser.

        Args:
            browser_name (str): Name of the browser to scan for extensions.

        Yields:

        .. code-block:: text

            Records with the following fields:
                ts_install (datetime): Extension install timestamp.
                ts_update (datetime): Extension update timestamp.
                browser (string): The browser from which the records are generated.
                id (string): Extension unique identifier.
                name (string): Name of the extension.
                short_name (string): Short name of the extension.
                default_title (string): Default title of the extension.
                description (string): Description of the extension.
                version (string): Version of the extension.
                ext_path (path): Relative path of the extension.
                from_webstore (boolean): Extension from webstore.
                permissions (string[]): Permissions of the extension.
                manifest (varint): Version of the extensions' manifest.
                source: (path): The source file of the download record.
        """
        ext_files = ["Preferences", "Secure Preferences"]
        for filename in ext_files:
            for user, json_file, content in self._iter_json(filename):
                try:
                    extensions = content.get("extensions").get("settings")

                    for extension_id, extension_data in extensions.items():
                        ts_install = extension_data.get("first_install_time") or extension_data.get("install_time")
                        ts_update = extension_data.get("last_update_time")
                        if ts_install:
                            ts_install = webkittimestamp(ts_install)
                        if ts_update:
                            ts_update = webkittimestamp(ts_update)

                        if ext_path := extension_data.get("path"):
                            ext_path = self.target.fs.path(ext_path)

                        manifest = extension_data.get("manifest")
                        if manifest:
                            name = manifest.get("name")
                            short_name = manifest.get("short_name")
                            description = manifest.get("description")
                            ext_version = manifest.get("version")
                            ext_permissions = manifest.get("permissions")
                            manifest_version = manifest.get("manifest_version")

                            if manifest.get("browser_action"):
                                default_title = manifest.get("browser_action").get("default_title")
                            else:
                                default_title = None

                        else:
                            name = None
                            short_name = None
                            default_title = None
                            description = None
                            ext_version = None
                            ext_permissions = None
                            manifest_version = None

                        yield self.BrowserExtensionRecord(
                            ts_install=ts_install,
                            ts_update=ts_update,
                            browser=browser_name,
                            extension_id=extension_id,
                            name=name,
                            short_name=short_name,
                            default_title=default_title,
                            description=description,
                            version=ext_version,
                            ext_path=ext_path,
                            from_webstore=extensions.get(extension_id).get("from_webstore"),
                            permissions=ext_permissions,
                            manifest_version=manifest_version,
                            source=json_file,
                            _target=self.target,
                            _user=user.user,
                        )
                except (AttributeError, KeyError) as e:  # noqa: PERF203
                    self.target.log.warning("No browser extensions found in: %s", json_file)
                    self.target.log.debug("", exc_info=e)

    def passwords(self, browser_name: str | None = None) -> Iterator[BrowserPasswordRecord]:
        """Return browser password records from Chromium browsers.

        Chromium on Linux has ``basic``, ``gnome`` and ``kwallet`` methods for password storage:
            - ``basic`` ciphertext prefixed with ``v10`` and encrypted with hard coded parameters.
            - ``gnome`` and ``kwallet`` ciphertext prefixed with ``v11`` which is not implemented (yet).

        Chromium on Windows uses DPAPI user encryption with varying methods of encryption (``v10`` and ``v20``).

        The SHA1 hash of the user's password or the plaintext password is required to decrypt passwords
        when dealing with encrypted passwords created with Chromium v80 (February 2020) and newer (``v10``).

        Supports decrypting Windows App Bound Encryption passwords from Google Chrome and Microsoft Edge (``v20``).

        You can supply a SHA1 hash or plaintext password using the keychain (``-Kv`` or ``-K``).

        References:
            - https://chromium.googlesource.com/chromium/src/+/master/docs/linux/password_storage.md
            - https://chromium.googlesource.com/chromium/src/+/master/components/os_crypt/sync/os_crypt_linux.cc#40
        """

        for user, db_file, db in self._iter_db("Login Data"):
            keys = {}

            if self.target.os == OperatingSystem.WINDOWS.value:
                try:
                    local_state_path = db_file.parent.parent.joinpath("Local State")
                    keys = self.decryption_keys(local_state_path, user.user.name)
                except ValueError as e:
                    self.target.log.warning("Failed to decrypt local state key: %s", e)
                    self.target.log.debug("", exc_info=e)

            if not db.table("logins"):
                continue

            notes = {}
            if table := db.table("password_notes"):
                for row in table.rows():
                    notes[str(row.parent_id)] = row.value

            for row in db.table("logins").rows():
                encrypted_password: bytes = row.password_value
                decrypted_password = None

                # Attempt to decrypt password
                try:
                    decrypted_password = self.decrypt_value(user, keys, encrypted_password)
                except Exception as e:
                    self.target.log.warning("Failed to decrypt %r Chromium password: %s", encrypted_password[0:3], e)
                    self.target.log.debug("", exc_info=e)

                # Attempt to decrypt notes
                encrypted_notes = None
                decrypted_notes = None
                if encrypted_notes := notes.get(str(row.id)):
                    try:
                        decrypted_notes = self.decrypt_value(user, keys, encrypted_notes)
                    except Exception as e:
                        self.target.log.warning("Failed to decrypt %r Chromium note: %s", encrypted_notes[0:3], e)
                        self.target.log.debug("", exc_info=e)

                yield self.BrowserPasswordRecord(
                    ts_created=webkittimestamp(row.date_created),
                    ts_last_used=webkittimestamp(row.date_last_used),
                    ts_last_changed=webkittimestamp(row.date_password_modified or 0),
                    browser=browser_name,
                    id=row.id,
                    url=row.origin_url,
                    encrypted_username=None,
                    encrypted_password=row.password_value,
                    decrypted_username=row.username_value,
                    decrypted_password=decrypted_password,
                    encrypted_notes=encrypted_notes,
                    decrypted_notes=decrypted_notes,
                    source=db_file,
                    _target=self.target,
                    _user=user.user,
                )

    def decryption_keys(self, local_state_path: TargetPath, username: str) -> ChromiumKeys:
        """Return decrypted Chromium ``os_crypt.encrypted_key``and ``os_crypt.app_bound_encrypted_key`` values.

        Used by :meth:`ChromiumMixin.passwords` and :meth:`ChromiumMixin.cookies` for Windows targets.

        References:
            - https://security.googleblog.com/2024/07/improving-security-of-chrome-cookies-on.html
            - https://github.com/chromium/chromium/tree/main/chrome/browser/os_crypt
            - https://github.com/chromium/chromium/tree/main/chrome/elevation_service
            - https://chromium.googlesource.com/chromium/src/+/refs/heads/main/components/os_crypt/sync/os_crypt_win.cc
        """
        keys = ChromiumKeys()

        if not local_state_path.exists():
            self.target.log.warning("File does not exist: %s", local_state_path)
            return keys

        try:
            local_state_conf = json.loads(local_state_path.read_text())
        except json.JSONDecodeError as e:
            self.target.log.warning("File does not contain valid JSON: %s in %s", e, local_state_path)
            return keys

        if "os_crypt" not in local_state_conf:
            self.target.log.warning(
                "File does not contain os_crypt, Chromium is likely older than v80: %s",
                local_state_path,
            )
            return keys

        if not self.target.has_function("dpapi"):
            self.target.log.warning(
                "Unable to decrypt Chromium os_crypt keys: DPAPI plugin is not compatible with target"
            )
            return keys

        if not HAS_CRYPTO:
            self.target.log.warning("Missing pycryptodome dependency for crypto operations")
            return keys

        # Windows Chromium version 80 > uses AES ``os_crypt.encrypted_key`` which can be decrypted with user DPAPI keys.
        # Reference: https://chromium.googlesource.com/chromium/src/+/refs/heads/main/components/os_crypt/sync/os_crypt_win.cc
        try:
            encrypted_key = base64.b64decode(local_state_conf["os_crypt"]["encrypted_key"])[5:]
            aes_key = self.target.dpapi.decrypt_user_blob(encrypted_key, username)
            self.target.log.info("Decrypted Chromium OS Crypt key: %r", aes_key)
            keys.os_crypt_key = aes_key

        except (KeyError, ValueError) as e:
            self.target.log.warning("Unable to decode Chromium os_crypt encrypted_key in %s: %s", local_state_path, e)
            self.target.log.debug("", exc_info=e)

        # Windows Google Chrome versions > 130 / 127 and Microsoft Edge >~ 130 use App Bound Protection
        # (``os_crypt.app_bound_encrypted_key``) which can be decrypted using System DPAPI -> User DPAPI
        # -> (optionally AES GCM if Google Chrome).
        # Reference: https://github.com/chromium/chromium/tree/main/chrome/elevation_service
        if b64_abe := local_state_conf["os_crypt"].get("app_bound_encrypted_key"):
            try:
                app_bound_encrypted_key = base64.b64decode(b64_abe)

                if (header := app_bound_encrypted_key[0:4]) != b"APPB":
                    self.target.log.warning(
                        "Encountered unexpected app bound encrypted key in %s: Invalid header %r",
                        local_state_path,
                        header,
                    )
                    return keys

                system_ciphertext = app_bound_encrypted_key[4:].strip(b"\x00")
                user_ciphertext = self.target.dpapi.decrypt_system_blob(system_ciphertext)
                plaintext = self.target.dpapi.decrypt_user_blob(user_ciphertext, username)

                s_plaintext = c_elevation.Envelope(plaintext)

                if s_plaintext.ciphertext_len == 0x20:
                    # Microsoft Edge just stores the 32 byte decryption key
                    self.target.log.info("Decrypted Microsoft Edge ABE key: %r", s_plaintext.ciphertext)
                    keys.app_bound_key = s_plaintext.ciphertext

                else:
                    # Google Chrome has encrypted the AES decryption key using either AES GCM or ChaCha20 Poly1305
                    # using a static key.
                    data = c_elevation.GoogleChromeCipher(s_plaintext.ciphertext)

                    if data.flag == 0x01:
                        key = bytes.fromhex("b31c6e241ac846728da9c1fac4936651cffb944d143ab816276bcc6da0284787")
                        cipher = AES.new(key=key, mode=AES.MODE_GCM, nonce=data.iv)

                    elif data.flag == 0x02:
                        key = bytes.fromhex("e98f37d7f4e1fa433d19304dc2258042090e2d1d7eea7670d41f738d08729660")
                        cipher = ChaCha20_Poly1305.new(key=key, nonce=data.iv)

                    else:
                        raise ValueError(f"Unsupported ElevationService key flag {data.flag!r}")  # noqa: TRY301

                    aes_key = cipher.decrypt_and_verify(data.ciphertext, data.mac_tag)

                    self.target.log.info("Decrypted Google Chrome ABE key: %r", aes_key)
                    keys.app_bound_key = aes_key

            except (KeyError, ValueError, EOFError) as e:
                self.target.log.warning(
                    "Unable to decode Chromium os_crypt app_bound_key in %s: %s", local_state_path, e
                )
                self.target.log.debug("", exc_info=e)

        return keys

    def decrypt_value(self, user: UserDetails, keys: ChromiumKeys, encrypted: bytes) -> bytes:
        """Attempt to decrypt the given encrypted bytes."""

        DECRYPT_MAP = {
            OperatingSystem.WINDOWS.value: {
                b"\x01\x00\x00": decrypt_dpapi,  # First three bytes of DPAPI blob signature.
                b"v10": decrypt_v10_windows,
                b"v20": decrypt_v20_windows,
            },
            OperatingSystem.UNIX.value: {
                b"v10": decrypt_v10_linux,
                b"v11": decrypt_v11_linux,
            },
            OperatingSystem.LINUX.value: {
                b"v10": decrypt_v10_linux,
                b"v11": decrypt_v11_linux,
            },
        }
        return DECRYPT_MAP.get(self.target.os, {}).get(encrypted[0:3], decrypt_unsupported)(
            self.target, user, keys, encrypted
        )


class ChromiumPlugin(ChromiumMixin, BrowserPlugin):
    """Chromium browser plugin."""

    __namespace__ = "chromium"

    DIRS = (
        # Linux
        ".config/chromium/Default",
        ".var/app/org.chromium.Chromium/config/chromium/Default",
        "snap/chromium/common/chromium/Default",
        # Windows
        "AppData/Local/Chromium/User Data/Default",
    )

    @export(record=ChromiumMixin.BrowserHistoryRecord)
    def history(self) -> Iterator[ChromiumMixin.BrowserHistoryRecord]:
        """Return browser history records for Chromium browser."""
        yield from super().history("chromium")

    @export(record=ChromiumMixin.BrowserCookieRecord)
    def cookies(self) -> Iterator[ChromiumMixin.BrowserCookieRecord]:
        """Return browser cookie records for Chromium browser."""
        yield from super().cookies("chromium")

    @export(record=ChromiumMixin.BrowserDownloadRecord)
    def downloads(self) -> Iterator[ChromiumMixin.BrowserDownloadRecord]:
        """Return browser download records for Chromium browser."""
        yield from super().downloads("chromium")

    @export(record=ChromiumMixin.BrowserExtensionRecord)
    def extensions(self) -> Iterator[ChromiumMixin.BrowserExtensionRecord]:
        """Return browser extension records for Chromium browser."""
        yield from super().extensions("chromium")

    @export(record=ChromiumMixin.BrowserPasswordRecord)
    def passwords(self) -> Iterator[ChromiumMixin.BrowserPasswordRecord]:
        """Return browser password records for Chromium browser."""
        yield from super().passwords("chromium")


@dataclass(eq=True)
class ChromiumKeys:
    """Contains decrypted Chromium-variant encryption keys."""

    os_crypt_key: bytes | None = None
    """Windows Chromium version 80 and up for ``b"v10"`` ciphertexts."""

    app_bound_key: bytes | None = None
    """Windows Google Chrome and Microsoft Edge versions 127 and up for ``b"v20"`` ciphertexts."""


def decrypt_v10_linux(
    target: Target, user: UserDetails, keys: ChromiumKeys, encrypted: bytes, *, hardcoded_key: str = "peanuts"
) -> bytes | None:
    """Decrypt a version 10 Linux ciphertext.

    ``v10`` ciphertexts are encrypted using a PBKDF2 key derivation of the static string ``peanuts`` or an empty string
    and salt ``saltysalt`` using AES CBC with an IV of ``0x20 * 16``. Padded using PKCS7.

    Args:
        ciphertext: The encrypted bytes.

    Returns:
        Decrypted password string.

    References:
        - https://chromium.googlesource.com/chromium/src/+/refs/heads/main/components/os_crypt/sync/os_crypt_linux.cc
    """

    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency for AES operation")

    ciphertext = encrypted[3:]

    salt = b"saltysalt"
    iv = b" " * 16

    key = PBKDF2(hardcoded_key, salt, 16, 1)
    cipher = AES.new(key, AES.MODE_CBC, IV=iv)

    decrypted = cipher.decrypt(ciphertext)

    # No MAC so no way to verify if decryption worked other than check if unpadding works.
    # If unpadding failed, we try the legacy-legacy hardcoded empty key.
    try:
        return unpad(decrypted, 16)
    except ValueError:
        if hardcoded_key == "peanuts":
            return decrypt_v10_linux(target, user, keys, encrypted, hardcoded_key="")
        raise ValueError("Decrypting Chromium V10 secret failed")


def decrypt_v10_windows(target: Target, user: UserDetails, keys: ChromiumKeys, encrypted: bytes) -> bytes | None:
    """Decrypt a version 10 Windows ciphertext using key ``os_crypt_key``.

    ``v10`` variant 2 (Windows-specific) ciphertexts can be decrypted using a derived AES GCM
    key called ``os_crypt_key`` stored in an encrypted form in ``Local State`` files.

    The IV is prepended to the ciphertext as described in the structure definition below.

    References:

        .. code-block::

            struct chrome_pass {
                byte signature[3] = 'v10';
                byte iv[12];
                byte ciphertext[EOF];
            }

    Args:
        encrypted: Ciphertext bytes.
        key: The encryption key.

    Returns:
        Decrypted password string.
    """

    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency for AES operation")

    if not keys.os_crypt_key:
        raise ValueError("No OS Crypt key available for v10 version 2 decryption")

    iv = encrypted[3:15]
    ciphertext = encrypted[15:]
    cipher = AES.new(key=keys.os_crypt_key, mode=AES.MODE_GCM, nonce=iv)
    plaintext = cipher.decrypt(ciphertext)
    return plaintext[:-16]


def decrypt_v20_windows(target: Target, user: UserDetails, keys: ChromiumKeys, encrypted: bytes) -> bytes | None:
    """Decrypt a version 20 ciphertext using App Bound Encryption (``app_bound_key``).

    ``v20`` (Windows) ciphertexts can be decrypted using a derived AES GCM key called ``app_bound_key``
    stored in a double or triple encrypted form in ``Local State`` files.

    The IV and a MAC-tag for verification are stored in the ciphertext blob as can be observed in the
    structure definition below.

    References:

        .. code-block::

            struct chrome_pass {
                byte flag[3] = 'v20';
                byte iv[12];
                byte ciphertext[...];
                byte mac_tag[16];
            };

    Args:
        data: Encrypted ciphertext in structured format with flag, iv, ciphertext and tag.
        key: AES GCM key to decrypt data with.

    Returns:
        Decrypted plaintext.
    """
    if not HAS_CRYPTO:
        raise ValueError("Missing pycryptodome dependency for AES operation")

    if not keys.app_bound_key:
        raise ValueError("No ABE key available for v20 decryption")

    iv = encrypted[3 : 3 + 12]
    ciphertext = encrypted[3 + 12 : -16]
    mac_tag = encrypted[-16:]

    cipher = AES.new(key=keys.app_bound_key, mode=AES.MODE_GCM, nonce=iv)
    return cipher.decrypt_and_verify(ciphertext, mac_tag)


def decrypt_v11_linux(target: Target, user: UserDetails, keys: ChromiumKeys, encrypted: bytes) -> None:
    """Decrypt Linux GNOME or Kwallet encrypted passwords. Currently not implemented."""
    raise ValueError("Decrypting v11 Linux GNOME or Kwallet Chromium ciphertexts is not implemented.")


def decrypt_dpapi(target: Target, user: UserDetails, keys: ChromiumKeys, encrypted: bytes) -> bytes | None:
    """Decrypt a DPAPI user blob for Windows-based Chromium installs.

    Chromium on Windows prior to version 80 encrypts passwords using user DPAPI master keys.

    They can be decrypted directly by utilizing the DPAPI plugin.

    References:
        - https://chromium.googlesource.com/chromium/src/+/refs/heads/main/components/os_crypt/sync/os_crypt_win.cc
    """

    if not target.has_function("dpapi"):
        raise ValueError("Missing DPAPI plugin for DPAPI user secret decryption")

    if not encrypted.startswith(b"\x01\x00\x00\x00\xd0\x8c\x9d\xdf"):
        raise ValueError("Provided blob does not look like a DPAPI Blob: {encrypted[:8]!r}")

    try:
        return target.dpapi.decrypt_user_blob(encrypted, user.user.name)

    except UnsupportedPluginError as e:
        target.log.warning("Target is missing required registry keys for DPAPI")
        target.log.debug("", exc_info=e)
    except ValueError as e:
        target.log.warning("Failed to decrypt DPAPI blob: %e", e)
        target.log.debug("", exc_info=e)


def decrypt_unsupported(target: Target, user: UserDetails, keys: ChromiumKeys, encrypted: bytes) -> None:
    raise ValueError(f"Unknown encrypted password found: {encrypted!r}")



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

UnixShadowRecord = TargetRecordDescriptor(
    "unix/shadow",
    [
        ("string", "name"),
        ("string", "crypt"),
        ("string", "salt"),
        ("string", "hash"),
        ("string", "algorithm"),
        ("string", "crypt_param"),
        ("datetime", "last_change"),
        ("datetime", "min_age"),
        ("datetime", "max_age"),
        ("varint", "warning_period"),
        ("varint", "inactivity_period"),
        ("datetime", "expiration_date"),
        ("string", "unused_field"),
    ],
)


class ShadowPlugin(Plugin):
    """Unix shadow passwords plugin."""

    SHADOW_FILES = ("/etc/shadow", "/etc/shadow-")

    def check_compatible(self) -> None:
        if not self.target.fs.path("/etc/shadow").exists():
            raise UnsupportedPluginError("No shadow file found")

    @export(record=UnixShadowRecord)
    def passwords(self) -> Iterator[UnixShadowRecord]:
        """Yield shadow records from /etc/shadow files.

        References:
            - https://manpages.ubuntu.com/manpages/oracular/en/man5/passwd.5.html
            - https://linux.die.net/man/5/shadow
        """

        seen_hashes = set()

        for shadow_file in self.SHADOW_FILES:
            if (path := self.target.fs.path(shadow_file)).exists():
                for line in path.open("rt"):
                    line = line.strip()
                    if line == "" or line.startswith("#"):
                        continue

                    shent = dict(enumerate(line.split(":")))
                    crypt = extract_crypt_details(shent)

                    # do not return a shadow record if we have no hash
                    if crypt.get("hash") is None or crypt.get("hash") == "":
                        continue

                    # prevent duplicate user hashes
                    current_hash = (shent.get(0), crypt.get("hash"))
                    if current_hash in seen_hashes:
                        continue

                    seen_hashes.add(current_hash)

                    # improve readability
                    last_change = None
                    min_age = None
                    max_age = None
                    expiration_date = None

                    try:
                        last_change = int(shent.get(2)) if shent.get(2) else None
                    except ValueError as e:
                        self.target.log.warning(
                            "Unable to parse last_change shadow value in %s: %s ('%s')", shadow_file, e, shent.get(2)
                        )

                    try:
                        min_age = int(shent.get(3)) if shent.get(3) else None
                    except ValueError as e:
                        self.target.log.warning(
                            "Unable to parse last_change shadow value in %s: %s ('%s')", shadow_file, e, shent.get(3)
                        )

                    try:
                        max_age = int(shent.get(4)) if shent.get(4) else None
                    except ValueError as e:
                        self.target.log.warning(
                            "Unable to parse last_change shadow value in %s: %s ('%s')", shadow_file, e, shent.get(4)
                        )

                    try:
                        expiration_date = int(shent.get(7)) if shent.get(7) else None
                    except ValueError as e:
                        self.target.log.warning(
                            "Unable to parse last_change shadow value in %s: %s ('%s')", shadow_file, e, shent.get(7)
                        )

                    yield UnixShadowRecord(
                        name=shent.get(0),
                        crypt=shent.get(1),
                        algorithm=crypt.get("algo"),
                        crypt_param=crypt.get("param"),
                        salt=crypt.get("salt"),
                        hash=crypt.get("hash"),
                        last_change=epoch_days_to_datetime(last_change) if last_change else None,
                        min_age=epoch_days_to_datetime(last_change + min_age) if last_change and min_age else None,
                        max_age=epoch_days_to_datetime(last_change + max_age) if last_change and max_age else None,
                        warning_period=shent.get(5) if shent.get(5) else None,
                        inactivity_period=shent.get(6) if shent.get(6) else None,
                        expiration_date=epoch_days_to_datetime(expiration_date) if expiration_date else None,
                        unused_field=shent.get(8),
                        _target=self.target,
                    )


def extract_crypt_details(shent: dict) -> dict:
    """Extract different parts of a shadow entry such as
    the used crypto algorithm, any parameters, the used salt and hash.
    """

    crypt = {"algo": None, "param": None, "salt": None, "hash": None}
    c_parts = shent.get(1).split("$")

    algos = {
        "$0$": "des",
        "$1$": "md5",
        "$2$": "bcrypt",
        "$2a$": "bcrypt",
        "$2b$": "bcrypt",
        "$2x$": "bcrypt",
        "$2y$": "eksbcrypt",
        "$5$": "sha256",
        "$6$": "sha512",
        "$y$": "yescrypt",
        "$gy$": "gost-yescrypt",
        "$7$": "scrypt",
    }

    # yescrypt and scrypt are structured as: $id$param$salt$hash
    if len(c_parts) == 5:
        crypt = {
            "algo": "$" + c_parts[1] + "$",
            "param": c_parts[2],
            "salt": c_parts[3],
            "hash": c_parts[4],
        }

    # others are usually structured as: $id$salt$hash
    elif len(c_parts) == 4:
        crypt = {
            "algo": "$" + c_parts[1] + "$",
            "param": None,
            "salt": c_parts[2],
            "hash": c_parts[3],
        }

    # display a nicer alrogrithm name
    if crypt["algo"] in algos:
        crypt["algo"] = algos[crypt["algo"]]

    return crypt


def epoch_days_to_datetime(days: int) -> datetime:
    """Convert a number representing the days since 1 January 1970 to a datetime object."""
    if not isinstance(days, int):
        raise TypeError("days argument should be an integer")

    return datetime(1970, 1, 1, 0, 0, tzinfo=timezone.utc) + timedelta(days)



from __future__ import annotations

import datetime
import itertools
import re
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.packagemanager import (
    OperationTypes,
    PackageManagerLogRecord,
    PackageManagerPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

APT_LOG_OPERATIONS = ["Install", "Reinstall", "Upgrade", "Downgrade", "Remove", "Purge"]
REGEX_PACKAGE_NAMES = re.compile(r"(.*?\)),?")


class AptPlugin(PackageManagerPlugin):
    """Apt package manager plugin."""

    __namespace__ = "apt"

    LOG_DIR_PATH = "/var/log/apt"
    LOG_FILES_GLOB = "history.*"

    def check_compatible(self) -> None:
        if not next(self.target.fs.path(self.LOG_DIR_PATH).glob(self.LOG_FILES_GLOB), None):
            raise UnsupportedPluginError("No APT files found")

    @export(record=PackageManagerLogRecord)
    def logs(self) -> Iterator[PackageManagerLogRecord]:
        """Package manager log parser for Apt.

        Apt creates logs that are multiline and therefore requires somewhat complex parsing logic.
        We create one ``PackageManagerLogRecord`` per package and type; the example below hence generates *three* records.

        Example log format::

            Start-Date: 2022-09-21  06:48:56
            Commandline: /usr/bin/unattended-upgrade
            Install: linux-headers-5.4.0-126:amd64 (5.4.0-126.142, automatic),
            Upgrade: linux-headers-generic:amd64 (5.4.0.125.126, 5.4.0.126.127), libpython3.9-minimal:amd64 (3.9.5-3ubuntu0~20.04.1, automatic)
            Requested-By: user (1000)
            End-Date: 2022-09-21  06:48:57
        """  # noqa E501
        target_tz = self.target.datetime.tzinfo

        for path in self.target.fs.path(self.LOG_DIR_PATH).glob(self.LOG_FILES_GLOB):
            chunk = []
            for line in itertools.chain(open_decompress(path, "rt"), [""]):
                line = line.strip()

                # Indicates the end of a log chunk
                if line == "":
                    yield from split_into_records(chunk, target_tz, self.target)

                    chunk = []
                    continue

                chunk.append(line)


def split_into_records(
    chunk: Iterator[str], tzinfo: datetime.tzinfo, target: Target
) -> Iterator[PackageManagerLogRecord]:
    """Parse the chunk line for line and try to extract as much information from each line as possible."""
    packages = []
    ts = None
    user = None
    command = None

    for line in chunk:
        if any(log_type in line for log_type in APT_LOG_OPERATIONS):
            operation, package_names = line.split(": ")
            package_names = split_package_names(package_names)

            packages.extend((operation, name) for name in package_names)

        elif line.startswith("Start-Date"):
            dt_string = line.split("Start-Date: ")[1]
            ts = datetime.datetime.strptime(dt_string, "%Y-%m-%d  %H:%M:%S").replace(tzinfo=tzinfo)

        elif line.startswith("Requested-By"):
            user = line.split("Requested-By: ")[1]

        elif line.startswith("Commandline: "):
            command = line.split("Commandline: ")[1]

    for operation, name in packages:
        yield PackageManagerLogRecord(
            ts=ts,
            package_manager="apt",
            package_name=name,
            operation=OperationTypes.infer(operation).value,
            requested_by_user=user,
            command=command,
            _target=target,
        )


def split_package_names(package_names: str) -> list[str]:
    """Splits a comma separated list of package names.

    Example ``package_names``::

        linux-headers-5.4.0-126:amd64 (5.4.0-126.142, automatic),
        linux-headers-5.4.0-126-generic:amd64 (5.4.0-126.142, automatic),
        linux-modules-extra-5.4.0-126-generic:amd64 (5.4.0-126.142, automatic),
        linux-modules-5.4.0-126-generic:amd64 (5.4.0-126.142, automatic),
        linux-image-5.4.0-126-generic:amd64 (5.4.0-126.142, automatic)

    Returns:
        A list of package names, e.g. ``['linux-headers-5.4.0-126:amd64 (5.4.0-126.142, automatic)', ...]``
    """

    package_names = REGEX_PACKAGE_NAMES.findall(package_names)
    return [name.strip() for name in package_names]



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from dissect.target import filesystem
from dissect.target.filesystems.tar import TarFilesystemDirectoryEntry, TarFilesystemEntry
from dissect.target.helpers import fsutil, loaderutil
from dissect.target.loaders.dir import find_and_map_dirs
from dissect.target.loaders.tar import TarSubLoader
from dissect.target.loaders.zip import ZipSubLoader

if TYPE_CHECKING:
    import tarfile as tf
    import zipfile as zf
    from pathlib import Path

    from dissect.target.target import Target


log = logging.getLogger(__name__)

FILESYSTEMS_ROOT = "fs"
FILESYSTEMS_LEGACY_ROOT = "sysvol"

ANON_FS_RE = re.compile(r"^fs[0-9]+$")


class AcquireTarSubLoader(TarSubLoader):
    """Loader for tar-based Acquire collections."""

    @staticmethod
    def detect(path: Path, tarfile: tf.TarFile) -> bool:
        for member in tarfile.getmembers():
            if member.name.startswith(
                (
                    f"/{FILESYSTEMS_ROOT}/",
                    f"{FILESYSTEMS_ROOT}/",
                    f"/{FILESYSTEMS_LEGACY_ROOT}/",
                    f"{FILESYSTEMS_LEGACY_ROOT}/",
                )
            ):
                return True
        return False

    def map(self, target: Target) -> None:
        volumes = {}

        for member in self.tar.getmembers():
            if member.name == ".":
                continue

            if member.name.startswith(("/fs/", "fs/")):
                # Current acquire
                parts = member.name.replace("fs/", "").split("/")
                if parts[0] == "":
                    parts.pop(0)
            else:
                # Legacy acquire
                parts = member.name.lstrip("/").split("/")
            volume_name = parts[0].lower()

            # NOTE: older versions of acquire would write to "sysvol" instead of a driver letter
            # Figuring out the sysvol from the drive letters is easier than the drive letter from "sysvol",
            # so this was swapped in acquire 3.12. Now we map all volumes to a drive letter and let the
            # Windows OS plugin figure out which is the sysvol
            # For backwards compatibility we're forced to keep this check, and assume that "c:" is our sysvol
            if volume_name == "sysvol":
                volume_name = "c:"

            if volume_name == "$fs$":
                if len(parts) == 1:
                    # The fs/$fs$ entry is ignored, only the directories below it are processed.
                    continue
                fs_name = parts[1]
                if ANON_FS_RE.match(fs_name):
                    parts.pop(0)
                    volume_name = f"{volume_name}/{fs_name}"

            if volume_name not in volumes:
                vol = filesystem.VirtualFilesystem(case_sensitive=False)
                vol.tar = self.tar
                volumes[volume_name] = vol
                target.filesystems.add(vol)

            volume = volumes[volume_name]
            mname = "/".join(parts[1:])

            entry_cls = TarFilesystemDirectoryEntry if member.isdir() else TarFilesystemEntry
            entry = entry_cls(volume, fsutil.normpath(mname), member)
            volume.map_file_entry(entry.path, entry)

        for vol_name, vol in volumes.items():
            loaderutil.add_virtual_ntfs_filesystem(
                target,
                vol,
                usnjrnl_path=[
                    "$Extend/$Usnjrnl:$J",
                    "$Extend/$Usnjrnl:J",  # Old versions of acquire used $Usnjrnl:J
                ],
            )

            target.fs.mount(vol_name, vol)


class AcquireZipSubLoader(ZipSubLoader):
    """Loader for zip-based Acquire collections."""

    @staticmethod
    def detect(path: Path, zipfile: zf.Path) -> bool:
        return zipfile.joinpath(FILESYSTEMS_ROOT).exists() or zipfile.joinpath(FILESYSTEMS_LEGACY_ROOT).exists()

    def map(self, target: Target) -> None:
        path = self.zip
        if path.joinpath(FILESYSTEMS_ROOT).exists():
            path = path.joinpath(FILESYSTEMS_ROOT)
        find_and_map_dirs(target, path)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable
from unittest.mock import patch

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.asdf import AsdfLoader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``AsdfLoader`` when opening a ``Target``."""
    path = absolute_path("_data/loaders/asdf/metadata.asdf")

    with patch("dissect.target.target.Target.apply"):
        target = opener(path)
        assert isinstance(target._loader, AsdfLoader)
        assert target.path == path


def test_loader_metadata() -> None:
    """Test the ASDF loader with metadata."""
    path = absolute_path("_data/loaders/asdf/metadata.asdf")

    loader = loader_open(path)
    assert isinstance(loader, AsdfLoader)

    t = Target()
    loader.map(t)
    assert len(t.filesystems) == 0

    assert list(map(str, t.fs.path("/").rglob("*"))) == [
        "/$asdf$",
        "/$asdf$/file_1",
        "/$asdf$/dir",
        "/$asdf$/dir/file_2",
    ]



"""A pathlib.Path compatible implementation for dissect.target.

This allows for the majority of the pathlib.Path API to "just work" on dissect.target filesystems.

Most of this consists of subclassed internal classes with dissect.target specific patches,
but sometimes the change to a function is small, so the entire internal function is copied
and only a small part changed. To ease updating this code, the order of functions, comments
and code style is kept largely the same as the original pathlib.py.

Yes, we know, this is playing with fire and it can break on new CPython releases.

The implementation is split up in multiple files, one for each CPython version.
You're currently looking at the CPython 3.11 implementation.

Commit hash we're in sync with: 846a23d

Notes:
    - CPython 3.11 ditched the _Accessor class, so we override the methods that should use it
"""

from __future__ import annotations

import fnmatch
import re
from pathlib import Path, PurePath, _PosixFlavour
from typing import IO, TYPE_CHECKING, Any, Callable, ClassVar

from dissect.target import filesystem
from dissect.target.exceptions import FilesystemError, SymlinkRecursionError
from dissect.target.helpers import polypath
from dissect.target.helpers.compat import path_common

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem, FilesystemEntry
    from dissect.target.helpers.fsutil import stat_result


class _DissectFlavour(_PosixFlavour):
    is_supported = True

    __variant_instances: ClassVar[dict[tuple[bool, str], _DissectFlavour]] = {}

    def __new__(cls, case_sensitive: bool = False, alt_separator: str = ""):
        idx = (case_sensitive, alt_separator)
        instance = cls.__variant_instances.get(idx, None)
        if instance is None:
            instance = _PosixFlavour.__new__(cls)
            cls.__variant_instances[idx] = instance

        return instance

    def __init__(self, case_sensitive: bool = False, alt_separator: str = ""):
        super().__init__()
        self.altsep = alt_separator
        self.case_sensitive = case_sensitive

    def casefold(self, s: str) -> str:
        return s if self.case_sensitive else s.lower()

    def casefold_parts(self, parts: list[str]) -> list[str]:
        return parts if self.case_sensitive else [p.lower() for p in parts]

    def compile_pattern(self, pattern: str) -> Callable[..., Any]:
        return re.compile(fnmatch.translate(pattern), 0 if self.case_sensitive else re.IGNORECASE).fullmatch

    def is_reserved(self, parts: list[str]) -> bool:
        return False


class PureDissectPath(PurePath):
    _fs: Filesystem
    _flavour = _DissectFlavour(case_sensitive=False)

    def __reduce__(self) -> tuple:
        raise TypeError("TargetPath pickling is currently not supported")

    @classmethod
    def _from_parts(cls, args: list) -> Self:
        fs = args[0]

        if not isinstance(fs, filesystem.Filesystem):
            raise TypeError(
                "invalid PureDissectPath initialization: missing filesystem, "
                f"got {args!r} (this might be a bug, please report)"
            )

        alt_separator = fs.alt_separator
        path_args = []
        for arg in args[1:]:
            if isinstance(arg, str):
                arg = polypath.normalize(arg, alt_separator=alt_separator)
            path_args.append(arg)

        self = super()._from_parts(path_args)
        self._fs = fs

        self._flavour = _DissectFlavour(alt_separator=fs.alt_separator, case_sensitive=fs.case_sensitive)

        return self

    def _make_child(self, args: list) -> Self:
        child = super()._make_child(args)
        child._fs = self._fs
        child._flavour = self._flavour
        return child

    def with_name(self, name: str) -> Self:
        result = super().with_name(name)
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    def with_stem(self, stem: str) -> Self:
        result = super().with_stem(stem)
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    def with_suffix(self, suffix: str) -> Self:
        result = super().with_suffix(suffix)
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    def relative_to(self, *other) -> Self:
        result = super().relative_to(*other)
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    def __rtruediv__(self, key: str) -> Self:
        try:
            return self._from_parts([self._fs, key, *self._parts])
        except TypeError:
            return NotImplemented

    @property
    def parent(self) -> Self:
        result = super().parent
        result._fs = self._fs
        result._flavour = self._flavour
        return result

    @property
    def parents(self) -> path_common._DissectPathParents:
        return path_common._DissectPathParents(self)


class TargetPath(Path, PureDissectPath):
    __slots__ = ("_entry",)

    def _make_child_relpath(self, part: str) -> Self:
        child = super()._make_child_relpath(part)
        child._fs = self._fs
        child._flavour = self._flavour
        return child

    def get(self) -> FilesystemEntry:
        try:
            return self._entry
        except AttributeError:
            self._entry = self._fs.get(str(self))
            return self._entry

    @classmethod
    def cwd(cls) -> Self:
        """Return a new path pointing to the current working directory
        (as returned by os.getcwd()).
        """
        raise NotImplementedError("TargetPath.cwd() is unsupported")

    @classmethod
    def home(cls) -> Self:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        raise NotImplementedError("TargetPath.home() is unsupported")

    def iterdir(self) -> Iterator[Self]:
        """Iterate over the files in this directory.  Does not yield any
        result for the special paths '.' and '..'.
        """
        for entry in path_common.scandir(self):
            if entry.name in {".", ".."}:
                # Yielding a path object for these makes little sense
                continue
            child_path = self._make_child_relpath(entry.name)
            child_path._entry = entry
            yield child_path

    def _scandir(self) -> path_common._DissectScandirIterator:
        return path_common.scandir(self)

    # NOTE: Forward compatibility with CPython >= 3.12
    def walk(
        self, top_down: bool = True, on_error: Callable[[Exception], None] | None = None, follow_symlinks: bool = False
    ) -> Iterator[tuple[Self, list[str], list[str]]]:
        """Walk the directory tree from this directory, similar to os.walk()."""
        paths = [self]

        while paths:
            path = paths.pop()
            if isinstance(path, tuple):
                yield path
                continue

            # We may not have read permission for self, in which case we can't
            # get a list of the files the directory contains. os.walk()
            # always suppressed the exception in that instance, rather than
            # blow up for a minor reason when (say) a thousand readable
            # directories are still left to visit. That logic is copied here.
            try:
                scandir_it = path._scandir()
            except OSError as e:
                if on_error is not None:
                    on_error(e)
                continue

            with scandir_it:
                dirnames = []
                filenames = []
                for entry in scandir_it:
                    try:
                        is_dir = entry.is_dir(follow_symlinks=follow_symlinks)
                    except OSError:
                        # Carried over from os.path.isdir().
                        is_dir = False

                    if is_dir:
                        dirnames.append(entry.name)
                    else:
                        filenames.append(entry.name)

            if top_down:
                yield path, dirnames, filenames
            else:
                paths.append((path, dirnames, filenames))

            paths += [path._make_child_relpath(d) for d in reversed(dirnames)]

    def absolute(self) -> Self:
        """Return an absolute version of this path.  This function works
        even if the path doesn't point to anything.

        No normalization is done, i.e. all '.' and '..' will be kept along.
        Use resolve() to get the canonical path to a file.
        """
        raise NotImplementedError("TargetPath.absolute() is unsupported in Dissect")

    # NOTE: We changed some of the error handling here to deal with our own exception types
    def resolve(self, strict: bool = False) -> Self:
        """
        Make the path absolute, resolving all symlinks on the way and also
        normalizing it.
        """

        s = path_common.realpath(self, strict=strict)
        p = self._from_parts((self._fs, s))

        # In non-strict mode, realpath() doesn't raise on symlink loops.
        # Ensure we get an exception by calling stat()
        if not strict:
            try:
                p.stat()
            except FilesystemError as e:
                if isinstance(e, SymlinkRecursionError):
                    raise
        return p

    def stat(self, *, follow_symlinks: bool = True) -> stat_result:
        """
        Return the result of the stat() system call on this path, like
        os.stat() does.
        """
        if follow_symlinks:
            return self.get().stat()
        return self.get().lstat()

    def owner(self) -> str:
        """
        Return the login name of the file owner.
        """
        raise NotImplementedError("TargetPath.owner() is unsupported")

    def group(self) -> str:
        """
        Return the group name of the file gid.
        """
        raise NotImplementedError("TargetPath.group() is unsupported")

    def open(
        self,
        mode: str = "rb",
        buffering: int = 0,
        encoding: str | None = None,
        errors: str | None = None,
        newline: str | None = None,
    ) -> IO:
        """Open file and return a stream.

        Supports a subset of features of the real pathlib.open/io.open.

        Note: in contrast to regular Python, the mode is binary by default. Text mode
        has to be explicitly specified. Buffering is also disabled by default.
        """
        return path_common.io_open(self, mode, buffering, encoding, errors, newline)

    def write_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        raise NotImplementedError("TargetPath.write_bytes() is unsupported")

    def write_text(
        self, data: str, encoding: str | None = None, errors: str | None = None, newline: str | None = None
    ) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        raise NotImplementedError("TargetPath.write_text() is unsupported")

    def readlink(self) -> Self:
        """
        Return the path to which the symbolic link points.
        """
        return self._from_parts((self._fs, self.get().readlink()))

    def touch(self, mode: int = 0o666, exist_ok: bool = True) -> None:
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        raise NotImplementedError("TargetPath.touch() is unsupported")

    def mkdir(self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        """
        Create a new directory at this given path.
        """
        raise NotImplementedError("TargetPath.mkdir() is unsupported")

    def chmod(self, mode: int, *, follow_symlinks: bool = True) -> None:
        """
        Change the permissions of the path, like os.chmod().
        """
        raise NotImplementedError("TargetPath.chmod() is unsupported")

    def lchmod(self, mode: int) -> None:
        """
        Like chmod(), except if the path points to a symlink, the symlink's
        permissions are changed, rather than its target's.
        """
        raise NotImplementedError("TargetPath.lchmod() is unsupported")

    def unlink(self, missing_ok: bool = False) -> None:
        """
        Remove this file or link.
        If the path is a directory, use rmdir() instead.
        """
        raise NotImplementedError("TargetPath.unlink() is unsupported")

    def rmdir(self) -> None:
        """
        Remove this directory.  The directory must be empty.
        """
        raise NotImplementedError("TargetPath.rmdir() is unsupported")

    def rename(self, target: str) -> Self:
        """
        Rename this path to the target path.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        raise NotImplementedError("TargetPath.rename() is unsupported")

    def replace(self, target: str) -> Self:
        """
        Rename this path to the target path, overwriting if that path exists.

        The target path may be absolute or relative. Relative paths are
        interpreted relative to the current working directory, *not* the
        directory of the Path object.

        Returns the new Path instance pointing to the target path.
        """
        raise NotImplementedError("TargetPath.replace() is unsupported")

    def symlink_to(self, target: str, target_is_directory: bool = False) -> None:
        """
        Make this path a symlink pointing to the target path.
        Note the order of arguments (link, target) is the reverse of os.symlink.
        """
        raise NotImplementedError("TargetPath.symlink_to() is unsupported")

    def hardlink_to(self, target: str) -> None:
        """
        Make this path a hard link pointing to the same file as *target*.

        Note the order of arguments (self, target) is the reverse of os.link's.
        """
        raise NotImplementedError("TargetPath.hardlink_to() is unsupported")

    def link_to(self, target: str) -> None:
        """
        Make the target path a hard link pointing to this path.

        Note this function does not make this path a hard link to *target*,
        despite the implication of the function and argument names. The order
        of arguments (target, link) is the reverse of Path.symlink_to, but
        matches that of os.link.

        Deprecated since Python 3.10 and scheduled for removal in Python 3.12.
        Use `hardlink_to()` instead.
        """
        raise NotImplementedError("TargetPath.link_to() is unsupported")

    def is_mount(self) -> bool:
        """
        Check if this path is a POSIX mount point
        """
        # Need to exist and be a dir
        if not self.exists() or not self.is_dir():
            return False

        try:
            parent_dev = self.parent.stat().st_dev
        except FilesystemError:
            return False

        dev = self.stat().st_dev
        if dev != parent_dev:
            return True
        ino = self.stat().st_ino
        parent_ino = self.parent.stat().st_ino
        return ino == parent_ino

    # NOTE: Forward compatibility with CPython >= 3.12
    def is_junction(self) -> bool:
        """
        Whether this path is a junction.
        """
        return path_common.isjunction(self)

    def expanduser(self) -> Self:
        """Return a new path with expanded ~ and ~user constructs
        (as returned by os.path.expanduser)
        """
        raise NotImplementedError("TargetPath.expanduser() is unsupported")



from __future__ import annotations

from dissect.target.plugin import NamespacePlugin, export

COMMON_EDITOR_FIELDS = [
    ("datetime", "ts"),
    ("string", "editor"),
    ("path", "source"),
]


class EditorPlugin(NamespacePlugin):
    """Editor plugin."""

    __namespace__ = "editor"

    @export
    def extensions(self) -> None:
        """Yields installed extensions."""
        raise NotImplementedError

    @export
    def history(self) -> None:
        """Yields history of files."""
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.windows.generic import UserRegistryRecordDescriptor

if TYPE_CHECKING:
    from collections.abc import Iterator

WinRarRecord = UserRegistryRecordDescriptor(
    "application/productivity/winrar",
    [
        ("datetime", "ts"),
        ("path", "path"),
    ],
)


class WinRarPlugin(Plugin):
    """Windows WinRAR GUI plugin."""

    BASE_KEY = "HKEY_CURRENT_USER\\Software\\WinRAR"

    def check_compatible(self) -> None:
        if not self.target.has_function("registry") or not list(self.target.registry.keys(self.BASE_KEY)):
            raise UnsupportedPluginError("No WinRAR registry keys found on target")

    @export(record=WinRarRecord)
    def winrar(self) -> Iterator[WinRarRecord]:
        """Return all available WinRAR history registry key values."""
        keys = [
            "ArcHistory",
            "DialogEditHistory\\ArcName",
            "DialogEditHistory\\ExtrPath",
        ]

        for key in keys:
            for r in self.target.registry.keys(f"{self.BASE_KEY}\\{key}"):
                user = self.target.registry.get_user(r)
                for v in r.values():
                    yield WinRarRecord(
                        ts=r.ts,
                        path=self.target.fs.path(v.value),
                        _key=r,
                        _user=user,
                        _target=self.target,
                    )



from __future__ import annotations

import hashlib
import hmac
from typing import TYPE_CHECKING

try:
    from Crypto.Cipher import AES, ARC4, DES3

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False

CIPHER_ALGORITHMS: dict[int | str, CipherAlgorithm] = {}
HASH_ALGORITHMS: dict[int | str, HashAlgorithm] = {}

if TYPE_CHECKING:
    from typing_extensions import Self


class CipherAlgorithm:
    id: int
    name: str
    key_length: int
    iv_length: int
    block_length: int

    def __init_subclass__(cls):
        CIPHER_ALGORITHMS[cls.id] = cls
        CIPHER_ALGORITHMS[cls.name] = cls

    @classmethod
    def from_id(cls, id: int) -> Self:
        return CIPHER_ALGORITHMS[id]()

    @classmethod
    def from_name(cls, name: str) -> Self:
        return CIPHER_ALGORITHMS[name]()

    def derive_key(self, key: bytes, hash_algorithm: HashAlgorithm) -> bytes:
        """Mimics the corresponding native Microsoft function.

        References:
            - https://github.com/tijldeneut/DPAPIck3/blob/main/dpapick3/crypto.py#L185
        """

        if len(key) > hash_algorithm.block_length:
            key = hashlib.new(hash_algorithm.name, key).digest()

        if len(key) >= self.key_length:
            return key

        key = key.ljust(hash_algorithm.block_length, b"\x00")
        pad1 = bytes(c ^ 0x36 for c in key)[: hash_algorithm.block_length]
        pad2 = bytes(c ^ 0x5C for c in key)[: hash_algorithm.block_length]
        key = hashlib.new(hash_algorithm.name, pad1).digest() + hashlib.new(hash_algorithm.name, pad2).digest()
        return self.fixup_key(key)

    def fixup_key(self, key: bytes) -> bytes:
        return key

    def decrypt_with_hmac(
        self, data: bytes, key: bytes, iv: bytes, hash_algorithm: HashAlgorithm, rounds: int
    ) -> bytes:
        derived = pbkdf2(key, iv, self.key_length + self.iv_length, rounds, hash_algorithm.name)
        key, iv = derived[: self.key_length], derived[self.key_length :]

        return self.decrypt(data, key, iv)

    def decrypt(self, data: bytes, key: bytes, iv: bytes | None = None) -> bytes:
        raise NotImplementedError


class _AES(CipherAlgorithm):
    id = 0x6611
    name = "AES"
    key_length = 128 // 8
    iv_length = 128 // 8
    block_length = 128 // 8

    def decrypt(self, data: bytes, key: bytes, iv: bytes | None = None) -> bytes:
        if not HAS_CRYPTO:
            raise RuntimeError("Missing pycryptodome dependency")

        cipher = AES.new(
            key[: self.key_length], mode=AES.MODE_CBC, IV=iv[: self.iv_length] if iv else b"\x00" * self.iv_length
        )
        return cipher.decrypt(data)


class _AES128(_AES):
    id = 0x660E
    name = "AES-128"


class _AES192(_AES):
    id = 0x660F
    name = "AES-192"
    key_length = 192 // 8


class _AES256(_AES):
    id = 0x6610
    name = "AES-256"
    key_length = 256 // 8


class _RC4(CipherAlgorithm):
    id = 0x6801
    name = "RC4"
    key_length = 40 // 8
    iv_length = 128 // 8
    block_length = 1 // 8

    def decrypt(self, data: bytes, key: bytes, iv: bytes | None = None) -> bytes:
        if not HAS_CRYPTO:
            raise RuntimeError("Missing pycryptodome dependency")

        cipher = ARC4.new(key[: self.key_length])
        return cipher.decrypt(data)


class _DES3(CipherAlgorithm):
    id = 0x6603
    name = "DES3"
    key_length = 192 // 8
    iv_length = 64 // 8
    block_length = 64 // 8

    def fixup_key(self, key: bytes) -> bytes:
        nkey = []
        for byte in key:
            parity_bit = 0
            for i in range(8):
                parity_bit ^= (byte >> i) & 1

            nkey.append(byte if parity_bit == 0 else byte | 1)
        return bytes(nkey[: self.key_length])

    def decrypt(self, data: bytes, key: bytes, iv: bytes | None = None) -> bytes:
        if not HAS_CRYPTO:
            raise RuntimeError("Missing pycryptodome dependency")

        if len(key) != 24:
            raise ValueError(f"Invalid DES3 CBC key length {len(key)}")

        cipher = DES3.new(key, DES3.MODE_CBC, iv=iv if iv else b"\x00" * 8)
        return cipher.decrypt(data)


class HashAlgorithm:
    id: int
    name: str
    digest_length: int
    block_length: int

    def __init_subclass__(cls):
        HASH_ALGORITHMS[cls.id] = cls
        HASH_ALGORITHMS[cls.name] = cls

    @classmethod
    def from_id(cls, id: int) -> Self:
        return HASH_ALGORITHMS[id]()

    @classmethod
    def from_name(cls, name: str) -> Self:
        return HASH_ALGORITHMS[name]()


class _MD5(HashAlgorithm):
    id = 0x8003
    name = "md5"
    digest_length = 128 // 8
    block_length = 512 // 8


class _SHA1(HashAlgorithm):
    id = 0x8004
    name = "sha1"
    digest_length = 160 // 8
    block_length = 512 // 8


class _HMAC(_SHA1):
    """Synonymous to SHA1."""

    id = 0x8009


class _SHA256(HashAlgorithm):
    id = 0x800C
    name = "sha256"
    digest_length = 256 // 8
    block_length = 512 // 8


class _SHA384(HashAlgorithm):
    id = 0x800D
    name = "sha384"
    digest_length = 384 // 8
    block_length = 1024 // 8


class _SHA512(HashAlgorithm):
    id = 0x800E
    name = "sha512"
    digest_length = 512 // 8
    block_length = 1024 // 8


def pbkdf2(passphrase: bytes, salt: bytes, key_len: int, iterations: int, digest: str = "sha1") -> bytes:
    """Implementation of PBKDF2 that allows specifying digest algorithm.

    Returns the corresponding expanded key which is ``key_len`` long.
    """
    key = bytearray()

    i = 1
    while len(key) < key_len:
        U = salt + i.to_bytes(4, "big")
        i += 1

        derived = hmac.new(passphrase, U, digestmod=digest).digest()
        for _ in range(iterations - 1):
            actual = hmac.new(passphrase, derived, digestmod=digest).digest()
            derived = bytes(x ^ y for x, y in zip(derived, actual))
        key.extend(derived)

    return bytes(key[:key_len])


def dpapi_hmac(pwd_hash: bytes, hmac_salt: bytes, value: bytes, hash_algorithm: HashAlgorithm) -> bytes:
    """Internal function used to compute HMACs of DPAPI structures."""
    key = hmac.new(pwd_hash, hmac_salt, digestmod=hash_algorithm.name).digest()
    return hmac.new(key, value, digestmod=hash_algorithm.name).digest()


def crypt_session_key_type1(
    master_key: bytes,
    nonce: bytes | None,
    hash_algorithm: HashAlgorithm,
    entropy: bytes | None = None,
    strong_password: str | None = None,
    smart_card_secret: bytes | None = None,
    verify_blob: bytes | None = None,
) -> bytes:
    """Computes the decryption key for Type1 DPAPI blob, given the master key and optional information.

    This implementation relies on a faulty implementation from Microsoft that does not respect the HMAC RFC.
    Instead of updating the inner pad, we update the outer pad.
    This algorithm is also used when checking the HMAC for integrity after decryption.

    Args:
        master_key: Decrypted master key (should be 64 bytes long).
        nonce: This is the nonce contained in the blob or the HMAC in the blob (integrity check).
        hash_algorithm: A :class:`HashAlgorithm` to use for calculating block sizes.
        entropy: This is the optional entropy from ``CryptProtectData()`` API.
        strong_password: Optional password used for decryption or the blob itself.
        smart_card_secret: Optional MS Next Gen Crypto secret (e.g. from PIN code).
        verify_blob: Optional encrypted blob used for integrity check.

    Returns:
        The decryption key.
    """
    if len(master_key) > 20:
        master_key = hashlib.sha1(master_key).digest()

    master_key = master_key.ljust(hash_algorithm.block_length, b"\x00")
    pad1 = bytes(c ^ 0x36 for c in master_key)
    pad2 = bytes(c ^ 0x5C for c in master_key)

    digest1 = hashlib.new(hash_algorithm.name)
    digest1.update(pad2)

    digest2 = hashlib.new(hash_algorithm.name)
    digest2.update(pad1)
    digest2.update(nonce)

    if entropy and smart_card_secret:
        digest2.update(entropy + smart_card_secret)
        if verify_blob:
            digest2.update(verify_blob)

    digest1.update(digest2.digest())
    if entropy and not smart_card_secret:
        digest1.update(entropy)

    if strong_password:
        strong_password = hashlib.sha1(strong_password.rstrip("\x00").encode("utf-16-le")).digest()
        digest1.update(strong_password)

    if verify_blob and not smart_card_secret:
        digest1.update(verify_blob)

    return digest1.digest()


def crypt_session_key_type2(
    masterkey: bytes,
    nonce: bytes,
    hash_algorithm: HashAlgorithm,
    entropy: bytes | None = None,
    strong_password: str | None = None,
    smart_card_secret: bytes | None = None,
    verify_blob: bytes | None = None,
) -> bytes:
    """Computes the decryption key for Type2 DPAPI blob, given the masterkey and optional information.

    This implementation relies on an RFC compliant HMAC implementation.
    This algorithm is also used when checking the HMAC for integrity after decryption.

    Args:
        master_key: Decrypted master key (should be 64 bytes long).
        nonce: This is the nonce contained in the blob or the HMAC in the blob (integrity check).
        hash_algo: A :class:`HashAlgorithm` to use for calculating block sizes.
        entropy: This is the optional entropy from ``CryptProtectData()`` API.
        strong_password: Optional password used for decryption or the blob itself.
        smart_card_secret: Optional MS Next Gen Crypto secret (e.g. from PIN code). Only for API compatibility.
        verify_blob: Optional encrypted blob used for integrity check.

    Returns:
        The decryption key.
    """
    if len(masterkey) > 20:
        masterkey = hashlib.sha1(masterkey).digest()

    digest = hmac.new(masterkey, digestmod=hash_algorithm.name)
    digest.update(nonce)

    if entropy:
        digest.update(entropy)

    if strong_password:
        strong_password = hashlib.sha512(strong_password.rstrip("\x00").encode("utf-16-le")).digest()
        digest.update(strong_password)

    elif verify_blob:
        digest.update(verify_blob)

    return digest.digest()


def derive_password_hash(password_hash: bytes, user_sid: str, digest: str = "sha1") -> bytes:
    """Internal use. Computes the encryption key from a user's password hash."""
    return hmac.new(password_hash, (user_sid + "\0").encode("utf-16-le"), digestmod=digest).digest()



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import RegistryKeyNotFoundError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import WindowsUserRecord, create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey

TrustedDocumentsRecord = create_extended_descriptor([RegistryRecordDescriptorExtension, UserRecordDescriptorExtension])(
    "windows/registry/trusteddocuments",
    [
        ("datetime", "ts_modified"),
        ("datetime", "ts_created"),
        ("datetime", "ts_enabled"),
        ("string", "application"),
        ("path", "document"),
        ("string", "state"),
        ("bytes", "raw"),
        ("path", "source"),
    ],
)

trustrecord_def = """
typedef QWORD FILETIME;

enum TRFLAG {
    EDITING_ENABLED     = 0x00000001
    MACROS_ENABLED      = 0x7fffffff
};

struct TrustRecordEntry {
    FILETIME            ts_created;
    QWORD               timezone_offset;
    DWORD               ts_enabled;
    TRFLAG              flag;
};
"""

c_trustrecord = cstruct().load(trustrecord_def)


def convert_time(time: int) -> float:
    """Return minute precision timestamp from DWORD time integer."""
    multiplier = 16505867190471736999  # E5109EC205D7BEA7
    return ((time << (64 + 29)) / multiplier) / 10_000_000


class TrustedDocumentsPlugin(Plugin):
    """Plugin to obtain Microsoft Office Trusted Document registry keys."""

    KEY = "HKEY_CURRENT_USER\\Software\\Microsoft\\Office"

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.key(self.KEY))) > 0:
            raise UnsupportedPluginError("No Trusted Document keys found")

    def _find_subkey(self, key: RegistryKey, subkey_name: str) -> Iterator[RegistryKey]:
        try:
            searched_key = key.subkey(subkey_name)
            if subkeys := searched_key.subkeys():
                yield from subkeys
            else:
                yield searched_key
        except RegistryKeyNotFoundError:
            pass

    def _iterate_keys(self) -> Iterator[tuple[WindowsUserRecord, str, RegistryKey]]:
        """Yields all Microsoft Office keys that contain a TrustRecords subkey."""
        for key in self.target.registry.keys(self.KEY):
            user = self.target.registry.get_user(key)
            for version_key in key.subkeys():
                for application_key in version_key.subkeys():
                    yield from (
                        (user, application_key.name, y)
                        for x in self._find_subkey(application_key, "Security")
                        for y in self._find_subkey(x, "TrustRecords")
                    )

    @export(record=TrustedDocumentsRecord)
    def trusteddocs(self) -> Iterator[TrustedDocumentsRecord]:
        """Return Microsoft Office TrustRecords registry keys for all Office applications.

        Microsoft uses Trusted Documents to cache whether the user enabled the editing and/or macros for that document.
        Therefore, this may reveal if macros have been enabled for a malicious Office document.

        Yields records based on the values within the TrustRecords registry keys.
        At least contains the following fields:

        .. code-block:: text

            application (string): Application name of the Office product that produced the TrustRecords registry key.
            document (path): Path to the document for which a TrustRecords entry is created.
            ts (datetime): The created time of the TrustRecord registry key.
            type (varint): Type of the value within the TrustRecords registry key.
            value (bytes): Value of the TrustRecords entry, which contains the information whether macros are enabled.

        References:
            - https://az4n6.blogspot.com/2016/02/more-on-trust-records-macros-and.html
            - https://github.com/DissectMalware/OfficeForensicTools/blob/master/trusted_documents.py
            - https://github.com/nmantani/PS-TrustedDocuments
        """
        for user, application, key in self._iterate_keys():
            for value in key.values():
                entry = c_trustrecord.TrustRecordEntry(value.value)
                yield TrustedDocumentsRecord(
                    ts_modified=key.ts,
                    ts_created=wintimestamp(entry.ts_created),
                    ts_enabled=convert_time(entry.ts_enabled),
                    application=application,
                    document=self.target.resolve(value.name),
                    state=entry.flag.name,
                    raw=value.value,
                    _key=key,
                    _user=user,
                    _target=self.target,
                )



from __future__ import annotations

import datetime
import math
import stat
from typing import TYPE_CHECKING, BinaryIO

from dissect.fat import exceptions as fat_exc
from dissect.fat import fat

from dissect.target.exceptions import FileNotFoundError, NotADirectoryError
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class FatFilesystem(Filesystem):
    __type__ = "fat"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, case_sensitive=False, alt_separator="\\", **kwargs)
        self.fatfs = fat.FATFS(fh)
        # FAT timestamps are in local time, so to prevent skewing them even more, we specify UTC by default.
        # However, it should be noted that they are not actual UTC timestamps!
        # Implementers can optionally set the tzinfo attribute of this class to get correct UTC timestamps.
        self.tzinfo = datetime.timezone.utc

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect a FAT filesystem on a given file-like object."""
        try:
            fat.validate_bpb(fh.read(512))
        except fat_exc.InvalidBPB:
            return False
        else:
            return True

    def get(self, path: str) -> FilesystemEntry:
        return FatFilesystemEntry(self, path, self._get_entry(path))

    def _get_entry(
        self, path: str, entry: fat.RootDirectory | fat.DirectoryEntry | None = None
    ) -> fat.RootDirectory | fat.DirectoryEntry:
        """Returns an internal FAT entry for a given path and optional relative entry."""
        try:
            return self.fatfs.get(path, dirent=entry)
        except fat_exc.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except fat_exc.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except fat_exc.Error as e:
            raise FileNotFoundError(path) from e


class FatFilesystemEntry(FilesystemEntry):
    def get(self, path: str) -> FilesystemEntry:
        """Get a filesystem entry relative from the current one."""
        full_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return FatFilesystemEntry(self.fs, full_path, self.fs._get_entry(path, self.entry))

    def open(self) -> BinaryIO:
        """Returns file handle (file-like object)."""
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self.entry.open()

    def iterdir(self) -> Iterator[str]:
        """List the directory contents of a directory. Returns a generator of strings."""
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for f in self.entry.iterdir():
            if f.name in (".", ".."):
                continue
            yield f.name

    def scandir(self) -> Iterator[FilesystemEntry]:
        """List the directory contents of this directory. Returns a generator of filesystem entries."""
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for f in self.entry.iterdir():
            if f.name in (".", ".."):
                continue
            path = fsutil.join(self.path, f.name, alt_separator=self.fs.alt_separator)
            yield FatFilesystemEntry(self.fs, path, f)

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return False

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        return self.entry.is_directory()

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        return not self.is_dir()

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                (stat.S_IFDIR if self.is_dir() else stat.S_IFREG) | 0o777,
                self.entry.cluster,
                id(self.fs),
                1,
                0,
                0,
                self.entry.size,
                self.entry.atime.replace(tzinfo=self.fs.tzinfo).timestamp(),
                self.entry.mtime.replace(tzinfo=self.fs.tzinfo).timestamp(),
                self.entry.ctime.replace(tzinfo=self.fs.tzinfo).timestamp(),
            ]
        )

        st_info.st_blocks = math.ceil(self.entry.size / self.entry.fs.cluster_size)
        st_info.st_blksize = self.entry.fs.cluster_size
        return st_info



from __future__ import annotations

from pathlib import Path


def absolute_path(filename: str) -> Path:
    return Path(__file__).parent.joinpath(filename).resolve()


def mkdirs(root: Path, paths: list[str]) -> None:
    for path in paths:
        root.joinpath(path).mkdir(parents=True)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.ntfs.c_ntfs import segment_reference

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.filesystem.ntfs.utils import get_drive_letter

if TYPE_CHECKING:
    from collections.abc import Iterator

UsnjrnlRecord = TargetRecordDescriptor(
    "filesystem/ntfs/usnjrnl",
    [
        ("datetime", "ts"),
        ("varint", "usn"),
        ("string", "segment_id"),
        ("path", "path"),
        ("string", "reason"),
        ("uint32", "security_id"),
        ("string", "source"),
        ("string", "attr"),
        ("uint16", "major"),
        ("uint16", "minor"),
    ],
)


class UsnjrnlPlugin(Plugin):
    """NFTS UsnJrnl plugin."""

    def check_compatible(self) -> None:
        if not any(fs for fs in self.target.filesystems if fs.__type__ == "ntfs"):
            raise UnsupportedPluginError("No NTFS filesystem(s) found on target")

    @export(record=UsnjrnlRecord)
    def usnjrnl(self) -> Iterator[UsnjrnlRecord]:
        """Return the UsnJrnl entries of all NTFS filesystems.

        The Update Sequence Number Journal (UsnJrnl) is a feature of an NTFS file system and contains information about
        filesystem activities. Each volume has its own UsnJrnl.

        If the filesystem is part of a virtual NTFS filesystem (a ``VirtualFilesystem`` with the UsnJrnl
        properties added to it through a "fake" ``NtfsFilesystem``), the paths returned in the UsnJrnl records
        are based on the mount point of the ``VirtualFilesystem``. This ensures that the proper original drive
        letter is used when available.
        When no drive letter can be determined, the path will show as e.g. ``\\$fs$\\fs0``.

        References:
            - https://en.wikipedia.org/wiki/USN_Journal
            - https://velociraptor.velocidex.com/the-windows-usn-journal-f0c55c9010e
        """
        target = self.target
        for fs in self.target.filesystems:
            if fs.__type__ != "ntfs":
                continue

            usnjrnl = fs.ntfs.usnjrnl
            if not usnjrnl:
                continue

            # If this filesystem is a "fake" NTFS filesystem, used to enhance a
            # VirtualFilesystem, The driveletter (more accurate mount point)
            # returned will be that of the VirtualFilesystem. This makes sure
            # the paths returned in the records are actually reachable.
            drive_letter = get_drive_letter(self.target, fs)
            for record in usnjrnl.records():
                try:
                    ts = None
                    try:
                        ts = record.timestamp
                    except ValueError as e:
                        target.log.error(  # noqa: TRY400
                            "Error occured during parsing of timestamp in usnjrnl: %x", record.record.TimeStamp
                        )
                        target.log.debug("", exc_info=e)

                    path = f"{drive_letter}{record.full_path}"
                    segment = segment_reference(record.record.FileReferenceNumber)
                    yield UsnjrnlRecord(
                        ts=ts,
                        usn=record.Usn,
                        segment_id=f"{segment}#{record.FileReferenceNumber.SequenceNumber}",
                        path=self.target.fs.path(path),
                        reason=str(record.Reason).replace("USN_REASON.", ""),
                        security_id=record.SecurityId,
                        source=str(record.SourceInfo).replace("USN_SOURCE.", ""),
                        attr=str(record.FileAttributes).replace("FILE_ATTRIBUTE.", ""),
                        major=record.MajorVersion,
                        minor=record.MinorVersion,
                        _target=target,
                    )
                except Exception as e:  # noqa: PERF203
                    target.log.error("Error during processing of usnjrnl record: %s", record.record)  # noqa: TRY400
                    target.log.debug("", exc_info=e)
